#include  <iostream>
#include  <cfloat>
using namespace std;
#include "error.hpp"
#include "AFunction.hpp"
#include "rgraph.hpp"
#include "RNM.hpp"
#include "fem.hpp"
#include "FESpace.hpp" 
#include "MeshPoint.hpp"
#include "BamgFreeFem.hpp"


#include "Operator.hpp"

#include <set>
#include <vector>
#include <fstream>

#include "lex.hpp"
#include "Mesh2.h"

#include "BamgFreeFem.hpp"
#include "lgfem.hpp"

// auxiliary functions
#include "cemracs2007aux.cpp"

using namespace Fem2D;

/**********************************/
/* good functions from the script*/

double rhs_reaction(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{ // return f-rp_h
  mp->set(x,y,0);
  return GetAny<double>( (*sol[3])(stack))-GetAny<double>( (*sol[7])(stack))*GetAny<double>( (*sol[0])(stack));
}

double reaction(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[7])(stack));
}

double ex(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[4])(stack));
}
double ex_dx(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[5])(stack));
}
double ex_dy(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[6])(stack));
}

double ph_dx(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[1])(stack));
}

double ph_dy(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[2])(stack));
}
/* end of good functios           */
/**********************************/

// PRAHA 29-10-2007 (Grenoble : moved here 3/11/07, because needed in compute_coef_2)
double compute_m_D_square(double C_F_or_C_P, double h_D, double c_r)
{
		  double m_D2_1 =  C_F_or_C_P*h_D*h_D;
		  double m_D2 = m_D2_1;
		  if (c_r>0) if (m_D2>1.0/c_r) m_D2=1.0/c_r;
		  return m_D2;
}
// PRAHA 29-10-2007  (Grenoble moved here 3/11/07)
double compute_trace_constant(double edge_length, double h_K, double area)
{
		return   3.0/2.0*edge_length*h_K/area;
}
// PRAHA 29-10-2007  (Grenoble moved here 3/11/07)
double compute_m_K(double C_F_or_C_P, double h_K, double c_r)
{
		  return sqrt(compute_m_D_square(C_F_or_C_P,h_K,c_r));
//		  double m_K = sqrt(C_F_or_C_P)*h_K;
//		  if (c_r > 0) { if (m_K>1.0/sqrt(c_r)) m_K = 1.0/sqrt(c_r); }
//		  return m_K;
}
// PRAHA 29-10-2007  (Grenoble moved here 3/11/07)
double compute_m_K_tilde(double h_K, double c_r)
{
		  double sqrtC_PK = 1.0/PI;
		  double m_K_tilde = (sqrtC_PK*sqrtC_PK+2.0/3.0*sqrtC_PK)*h_K;
		  double m_K_tilde_2 = 1.0/c_r*1.0/h_K + 1.0/3.0/sqrt(c_r);
		  if (m_K_tilde > m_K_tilde_2) m_K_tilde = m_K_tilde_2;
		  return m_K_tilde;
}

// TABOR 20071007
double qformula_exact_uni(double x1, double y1, double x2, double y2, double x3, double y3, MeshPoint* mp, vector<Expression> sol, Stack &stack )
{
  double aireK = 0.5*abs(x1*y3-x1*y2+x2*y1-x2*y3+x3*y2-x3*y1);
  double value=0, mx,my;
  // compute values of each midpoints
  mx = (x1+x2)/2.0; my = (y1+y2)/2.0;
  double vm1x = ph_dx(mx,my,mp,sol,stack) - ex_dx(mx,my,mp,sol,stack);
  double vm1y = ph_dy(mx,my,mp,sol,stack) - ex_dy(mx,my,mp,sol,stack);
  mx = (x2+x3)/2.0; my = (y2+y3)/2.0;
  double vm2x = ph_dx(mx,my,mp,sol,stack) - ex_dx(mx,my,mp,sol,stack);
  double vm2y = ph_dy(mx,my,mp,sol,stack) - ex_dy(mx,my,mp,sol,stack);
  mx = (x3+x1)/2.0; my = (y3+y1)/2.0;
  double vm3x = ph_dx(mx,my,mp,sol,stack) - ex_dx(mx,my,mp,sol,stack);
  double vm3y = ph_dy(mx,my,mp,sol,stack) - ex_dy(mx,my,mp,sol,stack);
  value = vm1x*vm1x + vm1y*vm1y + vm2x*vm2x + vm2y*vm2y + vm3x*vm3x + vm3y*vm3y;
  return (aireK/3.0*value); // ATTENTION : no sqrt
}

// TABOR 20071007
double qformula_DF(double x1, double y1, double x2, double y2, double x3, double y3, double alpha, double beta, double gamma, double gradphx, double gradphy, double aireK)
{
  double d= 2.0; // dimension of the problem
  double value=0;
  // compute midpoints
  double m1x = (x1+x2)/2.0;
  double m1y = (y1+y2)/2.0;
  double m2x = (x2+x3)/2.0;
  double m2y = (y2+y3)/2.0;
  double m3x = (x3+x1)/2.0;
  double m3y = (y3+y1)/2.0;
  // compute values of each midpoints
  double vm1x = gradphx + 1.0/(d*aireK)*(alpha*(m1x-x1) + beta*(m1x-x2) + gamma*(m1x-x3));
  double vm1y = gradphy + 1.0/(d*aireK)*(alpha*(m1y-y1) + beta*(m1y-y2) + gamma*(m1y-y3));
  double vm2x = gradphx + 1.0/(d*aireK)*(alpha*(m2x-x1) + beta*(m2x-x2) + gamma*(m2x-x3));
  double vm2y = gradphy + 1.0/(d*aireK)*(alpha*(m2y-y1) + beta*(m2y-y2) + gamma*(m2y-y3));
  double vm3x = gradphx + 1.0/(d*aireK)*(alpha*(m3x-x1) + beta*(m3x-x2) + gamma*(m3x-x3));
  double vm3y = gradphy + 1.0/(d*aireK)*(alpha*(m3y-y1) + beta*(m3y-y2) + gamma*(m3y-y3));
  value = vm1x*vm1x + vm1y*vm1y + vm2x*vm2x + vm2y*vm2y + vm3x*vm3x + vm3y*vm3y;
  return sqrt(aireK/3.0*value);
}

// TABOR 20071007
double qformula_R_uni_react(double x1, double y1, double x2, double y2, double x3, double y3, double alpha, double beta, double gamma, double aireK, MeshPoint* mp, vector<Expression> sol, Stack &stack)
{
  double value=0,my,mx;
  // compute midpoints
  // compute values of each midpoints
  mx = (x1+x2)/2.0;my = (y1+y2)/2.0;
  double vm1 = rhs_reaction(mx,my,mp,sol,stack) - (alpha + beta + gamma)/aireK;
  mx = (x2+x3)/2.0;my = (y2+y3)/2.0;
  double vm2 = rhs_reaction(mx,my,mp,sol,stack) - (alpha + beta + gamma)/aireK;
  mx = (x3+x1)/2.0;my = (y3+y1)/2.0;
  double vm3 = rhs_reaction(mx,my,mp,sol,stack) - (alpha + beta + gamma)/aireK;
  value = vm1*vm1 + vm2*vm2 + vm3*vm3;
  return sqrt(aireK/3.0*value);
}

// TABOR 20071007
double qformula_liberec(double x1, double y1, double x2, double y2, double x3, double y3, double aireK, MeshPoint* mp, vector<Expression> sol, Stack &stack)
// (f-r*ph,1)
{
  double value=0,my,mx;
  // compute midpoints
  // compute values of each midpoints
  mx = (x1+x2)/2.0;my = (y1+y2)/2.0;
  double vm1 = rhs_reaction(mx,my,mp,sol,stack);
  mx = (x2+x3)/2.0;my = (y2+y3)/2.0;
  double vm2 = rhs_reaction(mx,my,mp,sol,stack);
  mx = (x3+x1)/2.0;my = (y3+y1)/2.0;
  double vm3 = rhs_reaction(mx,my,mp,sol,stack);
  value = vm1 + vm2 + vm3;
  return (aireK/3.0*value);
}


// CEMRACS : tridiagonal + 2 elem solver
// RADEK 20-08-2007
void tridiag_solver(int N, double *a, double *b, double *c, double *f, double *y)
{
  if (N<2)
  {
    cout << "trouble tridiag_solver N = "<<N<<endl;
    return;
  }
  int i;
  double *w = new double[N]; 
  double *v = new double[N]; 
  double *z = new double[N]; 
  
  i=0; // initialization
  w[i] = a[i]; if (w[i]==0) cout << " CEMRACS  tridiag_solver : w["<<i<<"]="<<w[i]<<endl;
  v[i] = c[i]/w[i];
  z[i]= f[i]/w[i];
  for (i=1;i<N;i++)
  { 
    w[i] = a[i]-b[i]*v[i-1]; if (w[i]==0) cout << " CEMRACS  tridiag_solver : w["<<i<<"]="<<w[i]<<endl;
    v[i] = c[i]/w[i];
    z[i]= (f[i]-b[i]*z[i-1])/w[i];
  }
  y[N-1] = z[N-1];
  for (i=N-2;i>=0;i--) y[i] = z[i]-v[i]*y[i+1];
  delete [] w;
  delete [] v;
  delete [] z;
}


// TABOR 20071007
int cemracs_solver(int nu, double *alpha, double *beta, double *gamma, double *phi, double *upsilon)
// nu .. dimension of the problem C upsilon = phi
// alpha ... diagonal
// beta ... lower diagonal and beta[0] = C[N-1][0]
// gamma ... upper diagonal and gamma[N-1] = C[0][N-1]
{
  int singular=0;

  int N = nu-1;
  if (nu==1)
  {
    cout << "CEMRACS : cemracs_solver nu="<<nu <<endl;
    return 1;
  }
  // step 0: declaration : fill the matrix P = tridiag(N, a, b, c);
  double *a = new double[N]; for (int i=0;i<N;i++) a[i]=alpha[i];
  double *b = new double[N]; for (int i=0;i<N;i++) b[i]=beta[i];
  double *c = new double[N]; for (int i=0;i<N;i++) c[i]=gamma[i];
  double *f = new double[N]; for (int i=0;i<N;i++) f[i]=0;
  double *z1 = new double[N]; for (int i=0;i<N;i++) z1[i]=0;
  double *z2 =new double[N]; for (int i=0;i<N;i++) z2[i]=0;
  double *t = new double[N]; for (int i=0;i<N;i++) t[i]=0;
  double k1=0, k2=0,u=0;
  
  // 1.a
  for (int i=0;i<N;i++) f[i]=phi[i];
  tridiag_solver(N,a,b,c,f,z1);
  // 1.b
  k1 = beta[0]*z1[0] + beta[N]*z1[N-1];
  // 2.a
  for (int i=1;i<N-1;i++) f[i]=0;
  f[0] = gamma[nu-1]; f[N-1] = gamma[N-1];
  tridiag_solver(N,a,b,c,f,z2);
  // 2.b
  k2 = beta[0]*z2[0] + beta[N]*z2[N-1];
  // 3.
  if (fabs(-k2+alpha[nu-1])<=1e-5) 
  {
    cout << "CEMRACS cemracs_solve k2+a[nu-1][nu-1] = 0. close to a singular matrix. "<<-k2+a[nu-1] << " k2 "<< k2<< "a[ = " << alpha[nu-1] <<endl;
//    cout << " N = " << N << " diagonal: b   a   c    f"<<endl;
//    for (int j=0;j<nu;j++) cout << " b["<<j<<"]="<<beta[j] << " a["<<j<<"]="<<alpha[j] << " c["<<j<<"]="<<gamma[j]<< " f["<<j<<"]="<<phi[j] << endl;
    singular=1;
    u=0;
  } else
  u = (phi[nu-1]-k1)/(-k2+alpha[nu-1]);
  // 4. 
  for (int i=1;i<N-1;i++) f[i]=phi[i];
  f[0]=phi[0]-u*gamma[nu-1]; f[N-1]=phi[N-1]-u*gamma[N-1];
  tridiag_solver(N,a,b,c,f,t);
  // 5. set the solution
  for (int i=0;i<N;i++) upsilon[i]=t[i];
  upsilon[nu-1] = u;
 
  delete [] a;
  delete [] b;
  delete [] c;
  delete [] f;
  delete [] z1;
  delete [] z2;
  delete [] t;
  return singular;
}


// TABOR 20071007
void compute_matrix_int_tri(double *a,double *b,double *c,double *f, double **E_coef, int N)
// this completes the tridiagonal matrix
{
  for (int i=0; i<N-1;i++) a[i] = 2.0*(E_coef[i][4] + E_coef[i+1][5]);
  a[N-1] = 2.*(E_coef[N-1][4] + E_coef[0][5]);

  for (int i=0; i<N;i++) b[i] = E_coef[i][3];

  c[N-1] = E_coef[0][3];
  for (int i=0; i<N-1;i++) c[i] = E_coef[i+1][3];

  for (int i=0; i<N-1;i++) f[i] = -(E_coef[i][1] + E_coef[i+1][2]);
  f[N-1] = -(E_coef[N-1][1] + E_coef[0][2]);
}

// TABOR 20071007
void compute_matrix_ext_tri(double *a,double *b,double *c,double *f, double **E_coef, int N,
			double E10t, double E4n2t, double E1n2t)
// N = n-1 in paper
// this completes the tridiagonal matrix
{
  for (int i=0; i<N-1;i++) a[i] = 2.0*(E_coef[i][4] + E_coef[i+1][5]);
  a[N-1] = 2.*E4n2t;

  b[0]=0;
  for (int i=1; i<N;i++) b[i] = E_coef[i][3];
  
  c[N-1]=0;
  for (int i=0; i<N-1;i++) c[i] = E_coef[i+1][3];

  for (int i=1; i<N-1;i++) f[i] = -(E_coef[i][1] + E_coef[i+1][2]);
  f[N-1] =  -E1n2t;
  f[0] = -(E10t + E_coef[1][2]);
}


// TABOR 20071007
void compute_X0(double *X0, double vx, double vy, int N, double **A_triangle_coor, MeshPoint* mp, vector<Expression> sol, Stack &stack, int boundary)
{
  double x2, y2, n1, n2;
  if (boundary==0)
  {
    int ip1;
    for (int i=0;i<N;i++)
    {
      x2 = A_triangle_coor[i][2];
      y2 = A_triangle_coor[i][3];
      n1 = (vy-y2);
      n2 = -(vx-x2);
      ip1=(i<N-1) ? i+1 : 0;
      double bx, by;   		  
      barycenter(bx,by,vx,vy,A_triangle_coor[i][0],A_triangle_coor[i][1],A_triangle_coor[i][2],A_triangle_coor[i][3]);
      double gx1 = ph_dx(bx,by,mp,sol,stack);
      double gy1 = ph_dy(bx,by,mp,sol,stack);
      barycenter(bx,by,vx,vy,A_triangle_coor[ip1][0],A_triangle_coor[ip1][1],A_triangle_coor[ip1][2],A_triangle_coor[ip1][3]);
      double gx2 = ph_dx(bx,by,mp,sol,stack);
      double gy2 = ph_dy(bx,by,mp,sol,stack);
      X0[i] = -0.5*((gx1+gx2)*n1 + (gy1+gy2)*n2);
    }
  }
  if (boundary!=0)
  {
    int ip1;
    for (int i=0;i<N;i++)
    {
      x2 = A_triangle_coor[i][2];
      y2 = A_triangle_coor[i][3];
      n1 = (vy-y2);
      n2 = -(vx-x2);
      double bx, by;   		  
      barycenter(bx,by,vx,vy,A_triangle_coor[i][0],A_triangle_coor[i][1],A_triangle_coor[i][2],A_triangle_coor[i][3]);
      double gx1 = ph_dx(bx,by,mp,sol,stack);
      double gy1 = ph_dy(bx,by,mp,sol,stack);
      barycenter(bx,by,vx,vy,A_triangle_coor[i+1][0],A_triangle_coor[i+1][1],A_triangle_coor[i+1][2],A_triangle_coor[i+1][3]);
      double gx2 = ph_dx(bx,by,mp,sol,stack);
      double gy2 = ph_dy(bx,by,mp,sol,stack);
      X0[i] = -0.5*((gx1+gx2)*n1 + (gy1+gy2)*n2);
    }
  }
}


// CEMRACS min 3
// IBRAHIM 20-08-2007
// grenoble 3/11/07 : name compute_coef_reaction changed into compute_coef_1
void compute_coef_1(int i, double **Res_coef, double **Diff_coef, double **E_coef, double **alpha_tab, MeshPoint* mp, vector<Expression> sol, Stack &stack
		  ,double xx0, double yy0, double xx1, double yy1, double xx2, double yy2, double gx, double gy, double mD2)
// compute coeficients for  for eta_R+eta_DF1
{
  double d = 2.0;
  //coord of the vertices of the small triangle
  double area = 0.5*fabs(xx0*yy2-xx0*yy1+xx1*yy0-xx1*yy2+xx2*yy1-xx2*yy0);
  double n0x = (yy2-yy1);
  double n0y = (xx1-xx2);
  double alpha0 = -(gx*n0x + gy*n0y);
  double m0x = 0.5*(xx2+xx1);
  double m0y = 0.5*(yy2+yy1);
  double m1x = 0.5*(xx2+xx0);
  double m1y = 0.5*(yy2+yy0);
  double m2x = 0.5*(xx0+xx1);
  double m2y = 0.5*(yy0+yy1);
  double f0 = area*rhs_reaction(m0x, m0y, mp,sol,stack);//coef_de_f);MeshPoint* mp, vector<Expression> sol, Stack &stack
  double f1 = area*rhs_reaction(m1x, m1y, mp,sol,stack);//coef_de_f);
  double f2 = area*rhs_reaction(m2x, m2y, mp,sol,stack);//coef_de_f);
//  double C = mD2/(3.0*d*d*area);
  double C = mD2/(3.0*area);
  Res_coef[i][0] = C*(f0*f0+f1*f1+f2*f2-2.*(f0+f1+f2)*alpha0+3.*alpha0*alpha0);
  Res_coef[i][1] = C*(-2.*(f0+f1+f2)+6.*alpha0);
  Res_coef[i][2] = -C*(-2.*(f0+f1+f2)+6.*alpha0);
  Res_coef[i][3] = -6.*C;
  Res_coef[i][4] = 3.*C;
  Res_coef[i][5] = 3.*C;
  double t0x0 = alpha0*(m0x-xx0);
  double t1x0 = m0x-xx1;
  double t2x0 = -(m0x-xx2); 
  double t0y0 = alpha0*(m0y-yy0);
  double t1y0 = m0y-yy1;
  double t2y0 = -(m0y-yy2); 
  double t0x1 = alpha0*(m1x-xx0);
  double t1x1 = m1x-xx1;
  double t2x1 = -(m1x-xx2); 
  double t0y1 = alpha0*(m1y-yy0);
  double t1y1 = m1y-yy1;
  double t2y1 = -(m1y-yy2); 
  double t0x2 = alpha0*(m2x-xx0);
  double t1x2 = m2x-xx1;
  double t2x2 = -(m2x-xx2); 
  double t0y2 = alpha0*(m2y-yy0);
  double t1y2 = m2y-yy1;
  double t2y2 = -(m2y-yy2); 
  double D = 1./(3.0*d*d*area);
  Diff_coef[i][0] =   D*((d*area*gx+t0x0)*(d*area*gx+t0x0)+(d*area*gy+t0y0)*(d*area*gy+t0y0)
		       + (d*area*gx+t0x1)*(d*area*gx+t0x1)+(d*area*gy+t0y1)*(d*area*gy+t0y1)
		       + (d*area*gx+t0x2)*(d*area*gx+t0x2)+(d*area*gy+t0y2)*(d*area*gy+t0y2));
  Diff_coef[i][1] = D*(2.*((d*area*gx+t0x0)*t1x0+(d*area*gy+t0y0)*t1y0)
		       + 2.*((d*area*gx+t0x1)*t1x1+(d*area*gy+t0y1)*t1y1)
		       + 2.*((d*area*gx+t0x2)*t1x2+(d*area*gy+t0y2)*t1y2));
  Diff_coef[i][2] = D*(2.*((d*area*gx+t0x0)*t2x0+(d*area*gy+t0y0)*t2y0)
		       + 2.*((d*area*gx+t0x1)*t2x1+(d*area*gy+t0y1)*t2y1)
		       + 2.*((d*area*gx+t0x2)*t2x2+(d*area*gy+t0y2)*t2y2));
  Diff_coef[i][3] = 2.0*D*(t1x0*t2x0+t1y0*t2y0
		       + t1x1*t2x1+t1y1*t2y1
		       + t1x2*t2x2+t1y2*t2y2);
  Diff_coef[i][4] = D*(t1x0*t1x0 + t1y0*t1y0
		       + t1x1*t1x1 + t1y1*t1y1
		       + t1x2*t1x2 + t1y2*t1y2);
  Diff_coef[i][5] = D*(t2x0*t2x0 + t2y0*t2y0
		       + t2x1*t2x1 + t2y1*t2y1
		       + t2x2*t2x2 + t2y2*t2y2);
  for (int j = 0; j<6; j++)
    E_coef[i][j] = Res_coef[i][j] + Diff_coef[i][j];	
//  alpha_tab[i][0] = alpha0; // TODO : please Ibrahim, explain, no idead why using this stuff here !!!!
}


// Grenoble 3/11/07
void compute_coef_2(int i, double **Res_coef, double **Diff_coef, double **E_coef, double **alpha_tab, MeshPoint* mp, vector<Expression> sol, Stack &stack
		  ,double xx0, double yy0, double xx1, double yy1, double xx2, double yy2, double gx, double gy, double mD2, double C_F_or_C_P, double c_r)
// compute coeficients for eta_R+eta_DF2
{
  double d = 2.0;
  //eta_R
  double area = 0.5*fabs(xx0*yy2-xx0*yy1+xx1*yy0-xx1*yy2+xx2*yy1-xx2*yy0);
  double n0x = (yy2-yy1);
  double n0y = (xx1-xx2);
  double alpha0 = -(gx*n0x + gy*n0y);
  double m0x = 0.5*(xx2+xx1);
  double m0y = 0.5*(yy2+yy1);
  double m1x = 0.5*(xx2+xx0);
  double m1y = 0.5*(yy2+yy0);
  double m2x = 0.5*(xx0+xx1);
  double m2y = 0.5*(yy0+yy1);
  double f0 = area*rhs_reaction(m0x, m0y, mp,sol,stack);//coef_de_f);MeshPoint* mp, vector<Expression> sol, Stack &stack
  double f1 = area*rhs_reaction(m1x, m1y, mp,sol,stack);//coef_de_f);
  double f2 = area*rhs_reaction(m2x, m2y, mp,sol,stack);//coef_de_f);
//  double C = mD2/(3.0*d*d*area);
  double C = mD2/(3.0*area);
  Res_coef[i][0] = C*(f0*f0+f1*f1+f2*f2-2.*(f0+f1+f2)*alpha0+3.*alpha0*alpha0);
  Res_coef[i][1] = C*(-2.*(f0+f1+f2)+6.*alpha0);
  Res_coef[i][2] = -C*(-2.*(f0+f1+f2)+6.*alpha0);
  Res_coef[i][3] = -6.*C;
  Res_coef[i][4] = 3.*C;
  Res_coef[i][5] = 3.*C;
  //eta_DF2
  double h_K = compute_max_edge_triangle(xx0,yy0,xx1,yy1,xx2,yy2);
  double mK = compute_m_K(C_F_or_C_P, h_K, c_r);
  double mKt = compute_m_K_tilde(h_K, c_r);
  double s1 = sqrt((xx2-xx0)*(xx2-xx0)+(yy2-yy0)*(yy2-yy0));
  double s2 = sqrt((xx1-xx0)*(xx1-xx0)+(yy1-yy0)*(yy1-yy0));
  double Ct1 = compute_trace_constant(s1, h_K, area);
  double Ct2 = compute_trace_constant(s2, h_K, area);
  double n1x = (yy0-yy2)/s1;
  double n1y = (xx2-xx0)/s1;
  double n2x = (yy1-yy0)/s2;
  double n2y = (xx0-xx1)/s2;
  
  Diff_coef[i][0] = 2.*mK*mK*alpha0*alpha0/area + 4.*mKt*(Ct1*s1*(gx*n1x+gy*n1y)*(gx*n1x+gy*n1y)+  Ct2*s2*(gx*n2x+gy*n2y)*(gx*n2x+gy*n2y));
  Diff_coef[i][1] = 4.*mK*mK*alpha0/area + 8.*Ct1*mKt*(gx*n1x+gy*n1y);
  Diff_coef[i][2] = -4.*mK*mK*alpha0/area-8.*Ct2*mKt*(gx*n2x+gy*n2y);
  Diff_coef[i][3] = -4.*mK*mK/area;
  Diff_coef[i][4] = 2.*mK*mK/area+4.*Ct1*mKt/s1;
  Diff_coef[i][5] = 2.*mK*mK/area+4.*Ct2*mKt/s2;
  
  for (int j = 0; j<6; j++)
    E_coef[i][j] = Res_coef[i][j] + Diff_coef[i][j];	
}




// TABOR 20071007
void compute_grad_ph(int i, double &gx, double &gy, double vx, double vy, double **A_triangle_coor, MeshPoint* mp, vector<Expression> sol, Stack &stack)
{
	          double bx, by;   		  
		  barycenter(bx,by,vx,vy,A_triangle_coor[i][0],A_triangle_coor[i][1],A_triangle_coor[i][2],A_triangle_coor[i][3]);
		  gx = ph_dx(bx,by,mp,sol,stack);
		  gy = ph_dy(bx,by,mp,sol,stack);
}

// TABOR 20071007
double compute_estimators(double** alpha_tab, 
			double C_F_or_C_P,
			double **A_triangle_coor,
			double vx, double vy,
			int n_big_tria,
			MeshPoint* mp, vector<Expression> sol, Stack &stack, int boundary,
			double &est_DF,          // square root of the DF estimator
			double &est_R,		// square root of the R estimator
			double &est_DF_1,          // square root of the DF_1 estimator
			double &est_DF_2          // square root of the DF_2 estimator
			)           
{ // compute estimators
		double eta_DF_1=0,eta_DF_2=0; 
		double eta_R=0;
                    // compute the estimators for given alpha_tab // hard part !!!
 	        double h_D = compute_diameter(n_big_tria*2, A_triangle_coor, vx, vy);
                double c_r = reaction(vx,vy,mp,sol,stack); // TODO : this is not true if the reaction term is not constant... obvious
		for (int i=0 ; i<2*n_big_tria ; i++)
		{
		  double xx0 = vx;
		  double yy0 = vy;
		  double xx1 = A_triangle_coor[i][0];
		  double yy1 = A_triangle_coor[i][1];
		  double xx2 = A_triangle_coor[i][2];
		  double yy2 = A_triangle_coor[i][3];
		  double area = 0.5*abs(xx0*yy2-xx0*yy1+xx1*yy0-xx1*yy2+xx2*yy1-xx2*yy0);
		  
		  // compute eta_R
	          double gx, gy;   		  
		  compute_grad_ph(i,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		  
		  double m_D2 = compute_m_D_square(C_F_or_C_P,h_D,c_r);
		  double eta_R_loc = sqrt(m_D2)*qformula_R_uni_react(xx0, yy0, xx1, yy1, xx2, yy2, alpha_tab[i][0],alpha_tab[i][1], alpha_tab[i][2], area, mp, sol, stack);
		  eta_R += eta_R_loc*eta_R_loc;

		  // compute eta_DF_1
		  double eta_DF_1_sqrt = qformula_DF(xx0,yy0,xx1,yy1,xx2,yy2, alpha_tab[i][0],  alpha_tab[i][1], alpha_tab[i][2], gx, gy, area);
		  eta_DF_1 += eta_DF_1_sqrt*eta_DF_1_sqrt;
		  
		 if (c_r>0)
		 {
		  // compute eta_DF_2
		  // 1, find largest edge in the triangle
		  double h_K = compute_max_edge_triangle(xx0,yy0,xx1,yy1,xx2,yy2);
		  // 3. compute m_K
		  double m_K = compute_m_K(C_F_or_C_P,h_K,c_r);
		  // 4. compute m_K_tilde = trace_constant.pdf m_sigma
		  double m_K_tilde = compute_m_K_tilde(h_K,c_r);
		  // 5. compute norm || div t_h ||
		  double normdivth = fabs(alpha_tab[i][0]+alpha_tab[i][1]+alpha_tab[i][2])/sqrt(area);
		  // ext. normal to the edge nu 1 : n0
		  double n0x = (yy2-yy1);
		  double n0y = -(xx2-xx1);	
		  double sn0 = distance(xx1,yy1,xx2,yy2);
		  // ext. normal to the edge nu 1 : n1
		  double n1x = -(yy2-yy0);
		  double n1y = xx2-xx0;	
		  double sn1 = distance(xx0,yy0,xx2,yy2);
		  // ext. normal to the edge nu 2 : n2
		  double n2x =(yy1-yy0);
		  double n2y =-(xx1-xx0);
		  double sn2 = distance(xx0,yy0,xx1,yy1);
		  // compute grad ph inside a triangle
///		  double gx,gy;
//		  compute_grad_ph(i,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		  // now compute the sum of all edges of the triangle
//		  double j0=0,j1=0,j2=0;
      		  double j0 = gx*n0x + gy*n0y + alpha_tab[i][0];
		  double j1 = gx*n1x + gy*n1y + alpha_tab[i][1];
		  double j2 = gx*n2x + gy*n2y + alpha_tab[i][2];
		  //if (i==0 && boundary==1) j1 = 0;
		  //if (i==2*n_big_tria-1 && boundary==1) j2 = 0;
		  
		  double c_t_0 = compute_trace_constant(sn0,h_K,area);
		  double c_t_1 = compute_trace_constant(sn1,h_K,area);
		  double c_t_2 = compute_trace_constant(sn2,h_K,area);
		  // 5c. sum all of it
		  double sumofj = sqrt(c_t_0)*fabs(j0) + sqrt(c_t_1)*fabs(j1) + sqrt(c_t_2)*fabs(j2);
//		  if (j0 < 0 || j1 < 0 || j2 < 0) cout << "negative" << endl;
		  double eta_DF_2_r= m_K*normdivth + sqrt(m_K_tilde)*sumofj;
		  eta_DF_2 += eta_DF_2_r*eta_DF_2_r;
		  }
		}
		double eta_DF_min = (eta_DF_1 > eta_DF_2) ? eta_DF_2 : eta_DF_1;
		est_DF = (c_r>0) ?  sqrt(eta_DF_min) : sqrt(eta_DF_1);
//		est_DF=sqrt(eta_DF_1); // remove
		est_R = sqrt(eta_R);
		est_DF_1 = eta_DF_1;
		est_DF_2 = eta_DF_2;
		return (est_DF+est_R);
}

// TABOR-20071007
void compute_alpha_tab_from_X(double **alpha_tab, double alpha1, double alpha2, double *X, int n_big_tria, int boundary)
{
  	if (boundary == 0)
  	{
		int N = n_big_tria*2;
		alpha_tab[0][1] = X[0]; alpha_tab[0][2] = -X[N-1];
		for (int i=1 ; i<N ; i++) { alpha_tab[i][1] = X[i];  alpha_tab[i][2] = -X[i-1]; }
  	} else
  	{
//		  int N = n_big_tria*2-1;
		  alpha_tab[0][1] = X[0];
		  alpha_tab[0][2] = alpha2;
		  for (int i=1 ; i<n_big_tria*2-1 ; i++){
		    alpha_tab[i][1] = X[i];
		    alpha_tab[i][2] = -X[i-1];
		  }
		  alpha_tab[n_big_tria*2-1][1] = alpha1;
		  alpha_tab[n_big_tria*2-1][2] = -X[n_big_tria*2-2];		
 	}
}

// PRAHA 29-10-2007
void compute_alpha0(double **alpha_tab, int n_big_tria,  
			     double **A_triangle_coor,  double vx, double vy,
			     MeshPoint* mp, vector<Expression> sol, Stack &stack)
{
  for (int i=0;i<n_big_tria*2;i++)
  {
		double gx,gy;
		compute_grad_ph(i,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		double xx1 = A_triangle_coor[i][0];		
		double yy1 = A_triangle_coor[i][1];		
		double xx2 = A_triangle_coor[i][2];		
		double yy2 = A_triangle_coor[i][3];		
	        double n0x = (yy2-yy1);
		double n0y = (xx1-xx2);
	        alpha_tab[i][0] = -(gx*n0x + gy*n0y);
  }
}


// TABOR-20071007 (name modified in Grenoble 3/11/07 : ...minimization_1
void compute_th_minimization_1(double **alpha_tab, int n_big_tria,   
			     double **A_triangle_coor, double vx, double vy,
			     MeshPoint* mp, vector<Expression> sol, Stack &stack,
			     int boundary, double C_F_or_C_P
			     )
{
             compute_alpha0(alpha_tab,n_big_tria,A_triangle_coor,vx,vy,mp,sol,stack);
	      //coefficients of the polynomials
	      double **Res_coef;
	      double **Diff_coef;
	      double **E_coef;
	      Res_coef = new double* [n_big_tria*2];
	      Diff_coef = new double* [n_big_tria*2];
	      E_coef = new double* [n_big_tria*2];
	      for (int i=0;i<n_big_tria*2;i++){
		  Res_coef[i] = new double[6];
		  Diff_coef[i] = new double[6];
		  E_coef[i] = new double[6];
	     }
	     // compute h_D of the dual volume
	     double h_D = compute_diameter(n_big_tria*2, A_triangle_coor, vx, vy);
             double c_r = reaction(vx,vy,mp,sol,stack); // TODO : this is not true if the reaction term is not constant... obvious

	    // compute the residual and diff estimator coeficients for minimization
	      for (int i=0;i<n_big_tria*2;i++)
	      {
		double gx,gy;
		compute_grad_ph(i,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		double m_D2_1 =  C_F_or_C_P*h_D*h_D;
		double m_D2 = m_D2_1;
		if (c_r>0) if (m_D2>1.0/c_r) m_D2=1.0/c_r;
		compute_coef_1(i, Res_coef, Diff_coef, E_coef, alpha_tab, mp,sol,stack, vx, vy, A_triangle_coor[i][0], A_triangle_coor[i][1], A_triangle_coor[i][2], A_triangle_coor[i][3], gx, gy, m_D2);
	      }
	      //
 	      double alpha2=0, alpha1=0;
	       if (boundary==0)
	       {
		/////// matrix computation - analytical computation :
		int N = n_big_tria*2;
	        double *a = new double[N];
	        double *b = new double[N];
	        double *c = new double[N];
                double *f = new double[N];
	        double *X = new double[N]; // the solution

		compute_matrix_int_tri(a,b,c,f,E_coef, N); 
		int singular;
  		singular=cemracs_solver(N,a,b,c,f,X);

		compute_alpha_tab_from_X(alpha_tab, alpha1, alpha2, X, n_big_tria, boundary);
		/////// delete
		delete [] a;	
 		delete [] b;	
		delete [] c;			
		delete [] f;	
		delete [] X;	
	      } // END if boundary==0	      
	      if (boundary != 0)
	      {
		  //computation of alpha_2^0 :
		  double n2x = A_triangle_coor[0][1]-vy;
		  double n2y = vx-A_triangle_coor[0][0];
	          double gx, gy;   		  
	          compute_grad_ph(0,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		  alpha2 = -(gx*n2x + gy*n2y);
		  //computation of alpha_1^{n-1}
		  double n1x = vy-A_triangle_coor[n_big_tria*2-1][3];
		  double n1y = A_triangle_coor[n_big_tria*2-1][2]-vx;
	          
	          compute_grad_ph(n_big_tria*2-1,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		  alpha1 = -(gx*n1x + gy*n1y);
		  //computation of the "tilded" coef :
		  double E00t = E_coef[0][0] - E_coef[0][2]*alpha2 + E_coef[0][5]*alpha2*alpha2; 
		  double E10t = E_coef[0][1] - E_coef[0][3]*alpha2;
		  double E0n2t = E_coef[n_big_tria*2-2][0] + E_coef[n_big_tria*2-1][0]+ E_coef[n_big_tria*2-1][1]*alpha1 + E_coef[n_big_tria*2-1][4]*alpha1*alpha1;
		  double E1n2t = E_coef[n_big_tria*2-2][1] + E_coef[n_big_tria*2-1][2]+ E_coef[n_big_tria*2-1][3]*alpha1;
		  double E4n2t = E_coef[n_big_tria*2-2][4] + E_coef[n_big_tria*2-1][5];

		  /////// matrix computation///////
		  int N = n_big_tria*2-1;
	          double *a = new double[N];
	          double *b = new double[N];
	          double *c = new double[N];
                  double *f = new double[N];
	          double *X = new double[N]; 

		  compute_matrix_ext_tri(a,b,c,f,E_coef, N, E10t, E4n2t, E1n2t); 
		  if (N==1)
		  {
		    X[0] = -0.5*(E_coef[1][2]+E_coef[0][1]-E_coef[0][3]*alpha2+E_coef[1][3]*alpha1)/(E_coef[0][4]+E_coef[1][5]);
		  } else
		  {
		     tridiag_solver(N,a,b,c,f,X); // works better
		  } 
		  /////// fill alpha_tab ////////////////
 		  compute_alpha_tab_from_X(alpha_tab, alpha1, alpha2, X, n_big_tria, boundary);
		  ///////// delete 
		  delete [] a;	
 		  delete [] b;	
		  delete [] c;	
		  delete [] f;	
		  delete [] X;	
		} // end if boundary!=0
		// deletion
		for (int i=0 ; i<2*n_big_tria ; i++){
		  delete [] Res_coef[i];
		  delete [] Diff_coef[i];
		  delete [] E_coef[i];
		}
		delete [] Res_coef;
		delete [] Diff_coef;
		delete [] E_coef;
}


// Grenoble 3/11/2007
void compute_th_minimization_2(double **alpha_tab, int n_big_tria,   
			     double **A_triangle_coor, double vx, double vy,
			     MeshPoint* mp, vector<Expression> sol, Stack &stack,
			     int boundary, double C_F_or_C_P
			     )
{
             compute_alpha0(alpha_tab,n_big_tria,A_triangle_coor,vx,vy,mp,sol,stack);
	      //coefficients of the polynomials
	      double **Res_coef;
	      double **Diff_coef;
	      double **E_coef;
	      Res_coef = new double* [n_big_tria*2];
	      Diff_coef = new double* [n_big_tria*2];
	      E_coef = new double* [n_big_tria*2];
	      for (int i=0;i<n_big_tria*2;i++){
		  Res_coef[i] = new double[6];
		  Diff_coef[i] = new double[6];
		  E_coef[i] = new double[6];
	     }
	     // compute h_D of the dual volume
	     double h_D = compute_diameter(n_big_tria*2, A_triangle_coor, vx, vy);
             double c_r = reaction(vx,vy,mp,sol,stack); // TODO : this is not true if the reaction term is not constant... obvious

	    // compute the residual and diff estimator coeficients for minimization
	      for (int i=0;i<n_big_tria*2;i++)
	      {
		double gx,gy;
		compute_grad_ph(i,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		double m_D2_1 =  C_F_or_C_P*h_D*h_D;
		double m_D2 = m_D2_1;
		if (c_r>0) if (m_D2>1.0/c_r) m_D2=1.0/c_r;
		compute_coef_2(i, Res_coef, Diff_coef, E_coef, alpha_tab, mp,sol,stack, vx, vy, A_triangle_coor[i][0], A_triangle_coor[i][1], A_triangle_coor[i][2], A_triangle_coor[i][3], gx, gy, m_D2, C_F_or_C_P, c_r);
	      }
	      //
 	      double alpha2=0, alpha1=0;
	       if (boundary==0)
	       {
		/////// matrix computation - analytical computation :
		int N = n_big_tria*2;
	        double *a = new double[N];
	        double *b = new double[N];
	        double *c = new double[N];
                double *f = new double[N];
	        double *X = new double[N]; // the solution

		compute_matrix_int_tri(a,b,c,f,E_coef, N); 
		int singular;
  		singular=cemracs_solver(N,a,b,c,f,X);

		compute_alpha_tab_from_X(alpha_tab, alpha1, alpha2, X, n_big_tria, boundary);
		/////// delete
		delete [] a;	
 		delete [] b;	
		delete [] c;			
		delete [] f;	
		delete [] X;	
	      } // END if boundary==0	      
	      if (boundary != 0)
	      {
		  //computation of alpha_2^0 :
		  double n2x = A_triangle_coor[0][1]-vy;
		  double n2y = vx-A_triangle_coor[0][0];
	          double gx, gy;   		  
	          compute_grad_ph(0,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		  alpha2 = -(gx*n2x + gy*n2y);
		  //computation of alpha_1^{n-1}
		  double n1x = vy-A_triangle_coor[n_big_tria*2-1][3];
		  double n1y = A_triangle_coor[n_big_tria*2-1][2]-vx;
	          
	          compute_grad_ph(n_big_tria*2-1,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		  alpha1 = -(gx*n1x + gy*n1y);
		  //computation of the "tilded" coef :
		  double E00t = E_coef[0][0] - E_coef[0][2]*alpha2 + E_coef[0][5]*alpha2*alpha2; 
		  double E10t = E_coef[0][1] - E_coef[0][3]*alpha2;
		  double E0n2t = E_coef[n_big_tria*2-2][0] + E_coef[n_big_tria*2-1][0]+ E_coef[n_big_tria*2-1][1]*alpha1 + E_coef[n_big_tria*2-1][4]*alpha1*alpha1;
		  double E1n2t = E_coef[n_big_tria*2-2][1] + E_coef[n_big_tria*2-1][2]+ E_coef[n_big_tria*2-1][3]*alpha1;
		  double E4n2t = E_coef[n_big_tria*2-2][4] + E_coef[n_big_tria*2-1][5];

		  /////// matrix computation///////
		  int N = n_big_tria*2-1;
	          double *a = new double[N];
	          double *b = new double[N];
	          double *c = new double[N];
                  double *f = new double[N];
	          double *X = new double[N]; 

		  compute_matrix_ext_tri(a,b,c,f,E_coef, N, E10t, E4n2t, E1n2t); 
		  if (N==1)
		  {
		    X[0] = -0.5*(E_coef[1][2]+E_coef[0][1]-E_coef[0][3]*alpha2+E_coef[1][3]*alpha1)/(E_coef[0][4]+E_coef[1][5]);
		  } else
		  {
		     tridiag_solver(N,a,b,c,f,X); // works better
		  } 
		  /////// fill alpha_tab ////////////////
 		  compute_alpha_tab_from_X(alpha_tab, alpha1, alpha2, X, n_big_tria, boundary);
		  ///////// delete 
		  delete [] a;	
 		  delete [] b;	
		  delete [] c;	
		  delete [] f;	
		  delete [] X;	
		} // end if boundary!=0
		// deletion
		for (int i=0 ; i<2*n_big_tria ; i++){
		  delete [] Res_coef[i];
		  delete [] Diff_coef[i];
		  delete [] E_coef[i];
		}
		delete [] Res_coef;
		delete [] Diff_coef;
		delete [] E_coef;
}




// TABOR-20071007
void compute_th_jumping(double **alpha_tab, int n_big_tria,  
			     double **A_triangle_coor,  double vx, double vy,
			     MeshPoint* mp, vector<Expression> sol, Stack &stack,
			     int boundary, double C_F_or_C_P)
{
             compute_alpha0(alpha_tab,n_big_tria,A_triangle_coor,vx,vy,mp,sol,stack);
	      //
 	      double alpha2=0, alpha1=0;
	       if (boundary==0)
	       {
	        double *X = new double[n_big_tria*2]; 		
		compute_X0(X, vx, vy, n_big_tria*2, A_triangle_coor, mp, sol, stack, boundary);
		compute_alpha_tab_from_X(alpha_tab, alpha1, alpha2, X, n_big_tria, boundary);
		delete [] X;	
	      } // END if boundary==0	      
	      if (boundary != 0)
	      {
		  //computation of alpha_2^0 :
		  double n2x = A_triangle_coor[0][1]-vy;
		  double n2y = vx-A_triangle_coor[0][0];
	          double gx, gy;   		  
	          compute_grad_ph(0,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		  alpha2 = -(gx*n2x + gy*n2y);
		  //computation of alpha_1^{n-1}
		  double n1x = vy-A_triangle_coor[n_big_tria*2-1][3];
		  double n1y = A_triangle_coor[n_big_tria*2-1][2]-vx;
	          compute_grad_ph(n_big_tria*2-1,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		  alpha1 = -(gx*n1x + gy*n1y);
		  /////// matrix computation///////
	          double *X = new double[n_big_tria*2-1]; 
		  compute_X0(X, vx, vy, n_big_tria*2-1, A_triangle_coor, mp, sol, stack, boundary);
 		  compute_alpha_tab_from_X(alpha_tab, alpha1, alpha2, X, n_big_tria, boundary);
		  /////// fill alpha_tab ////////////////
		  delete [] X;	
		} // end if boundary!=0
}

// TABOR-20071007
void compute_th_liberec(double **alpha_tab, int n_big_tria,  
			     double **A_triangle_coor,  double vx, double vy,
			     MeshPoint* mp, vector<Expression> sol, Stack &stack,
			     int boundary, double C_F_or_C_P)
{
              compute_alpha0(alpha_tab,n_big_tria,A_triangle_coor,vx,vy,mp,sol,stack); // boundary fluxes
	      //
 	      double alpha2=0, alpha1=0;
	       if (boundary==0) // we are in the internal volumes
	       {
//	        double *X = new double[n_big_tria*2]; 		
//	           for (int i=0;i<N;i++)
		    int i = 0; // get the firt flux prescribed by the jump
                   {
                    double x2 = A_triangle_coor[i][2];
                    double y2 = A_triangle_coor[i][3];
                    double n1 = (vy-y2);
                    double n2 = -(vx-x2);
		      double bx, by;   		  
	              barycenter(bx,by,vx,vy,A_triangle_coor[i][0],A_triangle_coor[i][1],A_triangle_coor[i][2],A_triangle_coor[i][3]);
		      double gx1 = ph_dx(bx,by,mp,sol,stack);
    	              double gy1 = ph_dy(bx,by,mp,sol,stack);
		      barycenter(bx,by,vx,vy,A_triangle_coor[i+1][0],A_triangle_coor[i+1][1],A_triangle_coor[i+1][2],A_triangle_coor[i+1][3]);
		      double gx2 = ph_dx(bx,by,mp,sol,stack);
		      double gy2 = ph_dy(bx,by,mp,sol,stack);
	             alpha_tab[0][1] = -0.5*((gx1+gx2)*n1 + (gy1+gy2)*n2); // the jump est. in the first triangle
	            }
		    for (int j=1;j<n_big_tria*2;j++)
		    { 
		    //if (j > 0)  
		      double area = area_K(vx,vy,A_triangle_coor[j][0],A_triangle_coor[j][1],A_triangle_coor[j][2],A_triangle_coor[j][3]);
		      double sourceterm = qformula_liberec(vx,vy,A_triangle_coor[j][0],A_triangle_coor[j][1],A_triangle_coor[j][2],A_triangle_coor[j][3], area,mp,sol,stack);// integrate f-rph using q-formula
		    alpha_tab[j][2] = -alpha_tab[j-1][1];
		    alpha_tab[j][1] = -alpha_tab[j][2]-alpha_tab[j][0]+sourceterm;
//		      X[j] = X[j-1]-alpha_tab[j][0]+sourceterm;
		    }
		    alpha_tab[0][2]=-alpha_tab[n_big_tria*2-1][1];
		    
		    // ZKOUSKA
/**		    int j=0;
		    double area = area_K(vx,vy,A_triangle_coor[j][0],A_triangle_coor[j][1],A_triangle_coor[j][2],A_triangle_coor[j][3]);
		    double sourceterm = qformula_liberec(vx,vy,A_triangle_coor[j][0],A_triangle_coor[j][1],A_triangle_coor[j][2],A_triangle_coor[j][3], area,mp,sol,stack);// integrate f-rph using q-formula
//		    cout << "zkouska : " << alpha_tab[0][0]+alpha_tab[0][1]+alpha_tab[0][2] << " source : " << -sourceterm << " rozdil " 
//		    			<< alpha_tab[0][0]+alpha_tab[0][1]+alpha_tab[0][2]  -sourceterm << "area : " << area<< endl;
		    double diff=alpha_tab[0][0]+alpha_tab[0][1]+alpha_tab[0][2]  -sourceterm ;
		    if (fabs(diff)>1e-10) cout << "zkouska  : " << diff << endl;
/**/		    
		    
//		compute_X0(X, vx, vy, n_big_tria*2, A_triangle_coor, mp, sol, stack, boundary);
//		compute_alpha_tab_from_X(alpha_tab, alpha1, alpha2, X, n_big_tria, boundary);
//		delete [] X;	
	      } // END if boundary==0	      
	      if (boundary != 0)
	      {
		  //computation of alpha_2^0 :
		  double n2x = A_triangle_coor[0][1]-vy;
		  double n2y = vx-A_triangle_coor[0][0];
	          double gx, gy;   		  
	          compute_grad_ph(0,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		  alpha2 = -(gx*n2x + gy*n2y);
		  //computation of alpha_1^{n-1}
		  double n1x = vy-A_triangle_coor[n_big_tria*2-1][3];
		  double n1y = A_triangle_coor[n_big_tria*2-1][2]-vx;
	          compute_grad_ph(n_big_tria*2-1,gx,gy,vx,vy,A_triangle_coor,mp,sol,stack);
		  alpha1 = -(gx*n1x + gy*n1y);
		  /////// matrix computation///////
	          double *X = new double[n_big_tria*2-1]; 
		  alpha_tab[0][2] = alpha2;
		    for (int j=0;j<n_big_tria*2;j++)
		    { 
		      double area = area_K(vx,vy,A_triangle_coor[j][0],A_triangle_coor[j][1],A_triangle_coor[j][2],A_triangle_coor[j][3]);
		      double sourceterm = qformula_liberec(vx,vy,A_triangle_coor[j][0],A_triangle_coor[j][1],A_triangle_coor[j][2],A_triangle_coor[j][3], area,mp,sol,stack);// integrate f-rph using q-formula
//		      X[j] = X[j-1]-alpha_tab[j][0]+sourceterm;
		      if (j > 0)  
		         alpha_tab[j][2] = -alpha_tab[j-1][1];
		      alpha_tab[j][1] = -alpha_tab[j][2]-alpha_tab[j][0]+sourceterm;
		    }
//		    alpha1 = X[
//remove this - debug only :
//		  compute_X0(X, vx, vy, n_big_tria*2-1, A_triangle_coor, mp, sol, stack, boundary);
//		  compute_alpha_tab_from_X(alpha_tab, alpha1, alpha2, X, n_big_tria, boundary);
		  /////// fill alpha_tab ////////////////
		  delete [] X;	
		} // end if boundary!=0
}


// TABOR-20071007
/* This code is essential : computes  the estimator */
class AposterioCemracs2007 :   public E_F0mps { public:
  typedef double  Result;
  
  static basicAC_F0::name_and_type name_param[] ;
  static const int n_name_param =4;
  static const int n_output_param = 9;
  
  int nbsol;    
  Expression nargs[n_name_param];
  Expression getmesh;
  Expression em_output[n_output_param]; 
  // em_output[0] : eta_t ... eta_k
  // em_output[1] : eta_v
  // em_output[2] : eta_v_DF
  // em_output[3] : eta_v_R
  // em_output[4] : err_t ... err_k   = exact error per triangle
  // em_output[5] : err_v             = exact error per vertice = dual volume !
  // em_output[6] : h_d             = per vertex
  // em_output[7] : eta_v_DF_1      = per vertex
  // em_output[8] : eta_v_DF_2      = per vertex
  int  typesol[100];
  vector<Expression> sol;
  int nbcperiodic;
  Expression *periodic;
  
  
  double arg(int i,Stack stack,double a) const { return nargs[i] ? GetAny<double>( (*nargs[i])(stack) ): a;}
  long arg(int i,Stack stack,long a) const{ return nargs[i] ? GetAny<long>( (*nargs[i])(stack) ): a;}
  bool arg(int i,Stack stack,bool a) const{ return nargs[i] ? GetAny<bool>( (*nargs[i])(stack) ): a;}
  int arg(int i,Stack stack,int a) const{ return nargs[i] ? GetAny<int>( (*nargs[i])(stack) ): a;}
  
  AposterioCemracs2007(const basicAC_F0 & args) :nbsol(args.size()-1),sol(args.size()-1)
  {   
      args.SetNameParam(n_name_param,name_param,nargs);
      getmesh=to<pmesh>(args[0]); 
      int ksol=0; 
      ffassert(nbsol<100);
      for (int i=1;i<nbsol+1;i++)       
         if (args[i].left()==atype<E_Array>())
          {
            const E_Array * a = dynamic_cast<const E_Array *>(args[i].LeftValue());
            ffassert(a);
            ksol+=a->size(); 
          }
         else
           ksol++;
      sol.resize(ksol); 
      ksol=0; 
      for (int i=1;i<nbsol+1;i++)       
         if (args[i].left()==atype<E_Array>())
          {
            const E_Array * a = dynamic_cast<const E_Array *>(args[i].LeftValue());
            ffassert(a);
             int N=a->size();
            typesol[i-1]=N-1; // ok en 2D
            if (N<=8) { // used to be 4
                       for (int j=0;j<N;j++)             
                       sol[ksol++]=to<double>((*a)[j]); 
		     }
            else {
             lgerror(" AposterioCemracs2007 vecteur a plus de 4 compossantes inconnue");
            }
              
          }
         else {
          typesol[i-1]=0;
          sol[ksol++]=to<double>(args[i]);
          }
     const E_Array * expoutput = dynamic_cast<const E_Array *>(nargs[1]);
     if(expoutput)
     {
       if(expoutput->nbitem()!=n_output_param) ExecError("\nSorry we wait an array with 7 components in: output=[vector_triangles(P0) vector_vertices(P1) vector_eta_DF(P1) vector_eta_R(P1)  vector_err_k(P0) vector_err_v(P1) hd(P1) ]");
       for (int i=0;i<n_output_param;i++) em_output[i] = (*expoutput)[i];
     } else ExecError("\n CEMRACS_exec:  output=[] is missing ! ");
   }  



    static ArrayOfaType  typeargs() { return  ArrayOfaType(atype<pmesh>(),true);}
    static  E_F0 * f(const basicAC_F0 & args){ return new AposterioCemracs2007(args);} 
    AnyType operator()(Stack s) const ;    
    operator aType () const { return atype<pmesh>();}         
};

 basicAC_F0::name_and_type AposterioCemracs2007::name_param[AposterioCemracs2007::n_name_param] = {
       {   "medit",         &typeid(bool) }, // 0 do output in medit or not
       {   "output"          ,  &typeid(E_Array)},  // 1 arrays 
       {   "filename",         &typeid(string*)}, // 2 filename of data to be created (medit)
       {   "strategy",         &typeid(long) } // 3 adaptation strategy number, see later
    };



// TABOR-20071007
AnyType AposterioCemracs2007::operator()(Stack stack) const 
{
  using namespace bamg;
  double pi = bamg::Pi;
  double cutoffradian          = -1.0* pi/180. ;
  bool medit                   = arg(0,stack,false) ; // use medit for exporting data or not
  int strategy                 = arg(3,stack,0) ; // use medit for exporting data or not

  KN<double> *mm_output[n_output_param]; for (int i=0;i<n_output_param;i++) mm_output[i]=0;
  string*fn=0;
  if (nargs[2]) fn = GetAny<string*>((*nargs[2])(stack));
  using Fem2D::MeshPoint;
  using Fem2D::Mesh;
  Mesh * Thh = GetAny<pmesh>((*getmesh)(stack));
  ffassert(Thh);
  Triangles * oTh = msh2bamg(*Thh,cutoffradian);
  Triangles &Th(*oTh);

  Th.ReMakeTriangleContainingTheVertex();   // TODO : what does it do ? now idea...
  MeshPoint* mp(Fem2D::MeshPointStack(stack)); // ok - MeshPoint stack is reffered in mp
  long g_nt = Thh->nt; // the number of triangles
  long g_nv = Thh->nv; // the number of vertices

  for (int i=0;i<n_output_param;i++) mm_output[i]= GetAny<KN<double> *>( (*em_output[i])(stack) );
   
  KN<double> &vector_triangles=*mm_output[0];
  KN<double> &vector_vertices=*mm_output[1];
  KN<double> &vector_eta_DF=*mm_output[2];
  KN<double> &vector_eta_R=*mm_output[3];
  KN<double> &vector_err_t=*mm_output[4];
  KN<double> &vector_err_v=*mm_output[5];
  KN<double> &vector_h_d=*mm_output[6];
  KN<double> &vector_eta_DF_1=*mm_output[7];
  KN<double> &vector_eta_DF_2=*mm_output[8];

  // CEMRACS: output Dual mesh into a mesh file for visualisation of the error

  ofstream  of_mesh_eta_DF, of_mesh_eta_R, of_mesh_eta, of_mesh_err, of_mesh, of_mesh_exact;
  if (medit)
  {
    of_mesh.open((*fn+"_mesh.mesh").c_str()); // mesh file containing the 6-refined mesh 
    of_mesh_eta.open((*fn+"_mesh_eta.bb").c_str());//("_cemracs_eta_DF.bb"); // data file containinng the estimators 
    of_mesh_eta_DF.open((*fn+"_mesh_eta_DF.bb").c_str());//("_cemracs_eta_DF.bb"); // data file containinng the estimators 
    of_mesh_eta_R.open((*fn+"_mesh_eta_R.bb").c_str());//("_cemracs_eta_DF.bb"); // data file containinng the estimators 
    of_mesh_err.open((*fn+"_mesh_err.bb").c_str());//("_cemracs_eta_DF.bb"); // data file containinng the estimators 
    of_mesh_exact.open((*fn+"_mesh_exact.bb").c_str());//("_cemracs_eta_DF.bb"); // data file containinng the estimators 

    of_mesh << "MeshVersionFormatted 1" << endl;
    of_mesh << "Dimension 2" << endl;
    of_mesh << "Vertices " << (g_nt)*3 << endl;
  
    of_mesh_err    << "2 1 " << (g_nt)*3 << " 2 " << endl;
    of_mesh_eta    << "2 1 " << (g_nt)*3 << " 2 " << endl;
    of_mesh_eta_DF << "2 1 " << (g_nt)*3 << " 2 " << endl;
    of_mesh_eta_R  << "2 1 " << (g_nt)*3 << " 2 " << endl;
    of_mesh_exact  << "2 1 " << (g_nt)*3 << " 2 " << endl;
  } // end if (medit)

   // CEMRACS : go through all vertices and find all adjacent triangles
   int *aux_array = new int[g_nv]; // this is only an indicator used to be sure that we go through all vertices only once
   for (Int4 iv =0;iv<g_nv;iv++) aux_array[iv]=0;
    // go through all triangles and do algorithms for all vertices
    for (Int4  it = 0; it < g_nt; it++) 
    {
        bamg::Triangle & Kinit = Th(it); // current triangle we work on
	for (Int4 iv = 0; iv < 3; iv++)
	{
	  Int4 global_vertex_number = Th.Number(Kinit[iv]); // vertex we are messing around
    	  if (aux_array[global_vertex_number]==0)
	  {
	    // STEP -1 : initialization
  	    double vx = Kinit[iv].r.x;
	    double vy = Kinit[iv].r.y;
	    Int4 numKinit = it;               // initial number of the triangle we started from
	    Int4 numKcur =  numKinit;         // current number of the triangle we are in
	    Int4 done =0;                     // is the while loop over ?
	    Int4 boundary = 0; // are we on the boundary ?
	    Int4 numKstart = -1;
	  
	    Int4 num_adj_edge = 0; // number of the edge the next triangle is adjacent to 
	    Int4 local_vertex_num = iv;
	    Int4 n_big_tria = 0;   // number of big triangles around the vertex. 0 = there is no adjacent triangles
	    // STEP 0: now put information that this vertex was already computed
	    aux_array[global_vertex_number] = 1;
	    // STEP 1a: now go through the triangles  clock wise to find the boundary or simply count the global number of big triangles
	    while (done==0)
	    {
	      // now find the local index of the vertex in the numKcur triangle
	      n_big_tria++;
	      local_vertex_num=-1;
	      for (int i=0;i<3;i++) if (Th.Number(Th(numKcur)[i])==global_vertex_number) local_vertex_num = i;
	      if (local_vertex_num == -1) cout << "CEMRACS error : no match in the local vertex number : the vertex is not in the triangle !!! panic ..." << endl;
	      num_adj_edge = (local_vertex_num+2)%3;  // number of the edge for the adjacent triangle to go : clockwse
	      if (Th.Number(*Th(numKcur).Adj(num_adj_edge)) >= g_nt)  // we reached the edge from the left clock wise direction !!!
	      {
	         boundary = 1;
		 done = 1;
		 numKstart = numKcur;
	      } 
	      numKcur = Th.Number(*Th(numKcur).Adj(num_adj_edge));
	      if (numKcur == numKinit) // hooray, we meet the initial triangle again
	      {
	        done = 1;
		boundary = 0;
		numKstart = numKinit;
	      }
	    }
	    // STEP 1b : now adjust the number of triangles in the case of boundary : go anti-clock wise from initial triangle and only count the rest of the triangles
	    if (boundary==1)
	    {
	    n_big_tria--;
	    numKcur=numKinit;
	    boundary=0;
	    done=0;
	    while (done==0)
  	      {
	      // now find the local index of the vertex in the numKcur triangle
	      n_big_tria++;
	      local_vertex_num=-1;
	      for (int i=0;i<3;i++) if (Th.Number(Th(numKcur)[i])==global_vertex_number) local_vertex_num = i;
	      if (local_vertex_num == -1) cout << "CEMRACS error : no match in the local vertex number : the vertex is not in the triangle !!! panic ..." << endl;
	      num_adj_edge = (local_vertex_num+1)%3;  // number of the edge for the adjacent triangle to go : anticlockwise
	      if (Th.Number(*Th(numKcur).Adj(num_adj_edge)) >= g_nt)  // we reached the edge from the left clock wise direction !!!
	      {
	         boundary = 1;
		 done = 1;
	      } 
	      numKcur = Th.Number(*Th(numKcur).Adj(num_adj_edge));
	      if (numKcur == numKinit) // hooray, we meet the initial triangle again
	      {
	        done = 1;
		boundary = 0;
	      }
	      } // end while (done==0)
  	    if (boundary==0) cout << "CEMRACS critical error : no boundary detected in the 2nd run !!!" << endl;
	    }
	    
	    // STEP 2 : create the desired field of the small triangles 
	    // now go through all triangles from left to right and put data into the fields
	    double **A_triangle_coor = new double*[n_big_tria*2]; // field of small triangle vertex coordinates
	    for (int i=0;i<n_big_tria*2;i++) A_triangle_coor[i] = new double[4];  
	    done = 0;
	    int n_small_tria=0, n_tria_check=0;
	    numKcur=numKstart; // just start the loop from the most left positioned triangle (if boundary), go anticlockwise
	    while (done==0)
	    {
	      n_tria_check++;
	      if (n_tria_check > n_big_tria) cout << "CEMRACS error : number of triangles exceeded !!! " << endl;
	      // now find the local index of the vertex in the numKcur triangle
	      local_vertex_num=-1;
	      for (int i=0;i<3;i++) if (Th.Number(Th(numKcur)[i])==global_vertex_number) local_vertex_num = i;
	      if (local_vertex_num == -1) cout << "CEMRACS error : no match in the local vertex number : the vertex is not in the triangle !!! panic ..." << endl;
	      // a0. now set x1,y1, x2,x2, x0,y0
	         double x0 = Th(numKcur)[local_vertex_num].r.x;
		 double y0 = Th(numKcur)[local_vertex_num].r.y;
		 if (x0!=vx || y0!=vy) cout << "CEMRACS error : the vertex coordinates does not match local_vertex_num "<<endl;
	         double x1 = Th(numKcur)[(local_vertex_num+1)%3].r.x;
		 double y1 = Th(numKcur)[(local_vertex_num+1)%3].r.y;
	         double x2 = Th(numKcur)[(local_vertex_num+2)%3].r.x;
		 double y2 = Th(numKcur)[(local_vertex_num+2)%3].r.y;
	      // aa. now create 1st of 2 triangles
		 A_triangle_coor[n_small_tria][0] = (x0+x1)/2.0;
		 A_triangle_coor[n_small_tria][1] = (y0+y1)/2.0;
		 A_triangle_coor[n_small_tria][2] = (x0+x1+x2)/3.0;
		 A_triangle_coor[n_small_tria][3] = (y0+y1+y2)/3.0;
		 n_small_tria++;
		 if (n_small_tria>n_big_tria*2) cout << "CEMRACS error :aa n_small_tria>n_tria*2 ...nst " << n_small_tria << " nt " << n_big_tria  << endl;
	      // ab. now create 2nd of 2 triangles
		 A_triangle_coor[n_small_tria][0] = (x0+x1+x2)/3.0;
		 A_triangle_coor[n_small_tria][1] = (y0+y1+y2)/3.0;
		 A_triangle_coor[n_small_tria][2] = (x0+x2)/2.0;
		 A_triangle_coor[n_small_tria][3] = (y0+y2)/2.0;
		 n_small_tria++;
		 if (n_small_tria>n_big_tria*2) cout << "CEMRACS error :ab n_small_tria>n_tria*2 ...nst " << n_small_tria << " nt " << n_big_tria  << endl;
	      // now go to the next triangle
	      if (n_tria_check==n_big_tria) done =1;
	      if (done==0)
	      {
	      num_adj_edge = (local_vertex_num+1)%3;  // number of the edge for the adjacent triangle to go : anticlockwse
	      if (Th.Number(*Th(numKcur).Adj(num_adj_edge)) >= g_nt)  // we reached the edge from the left clock wise direction !!!
		 cout << "CEMRACS critical error : reached boundary but was not supposed to !!!!" << endl;
	      numKcur = Th.Number(*Th(numKcur).Adj(num_adj_edge));
	      if (numKcur == numKstart) // hooray, we meet the initial triangle again
		 cout << "CEMRACS critical error : reached the initial triangle but was not supposed to !!!!" << endl;
	      }
	    }
	// STEP 3 : computation of alpha - the degrees of freedom in the dual volume
	 double **alpha_tab_min = new double* [n_big_tria*2];
	 for (int i=0;i<2*n_big_tria;i++) alpha_tab_min[i]=new double[3];
	 double **alpha_tab_jump = new double* [n_big_tria*2];
	 for (int i=0;i<2*n_big_tria;i++) alpha_tab_jump[i]=new double[3];
	 double **alpha_tab_min_R = new double* [n_big_tria*2];
	 for (int i=0;i<2*n_big_tria;i++) alpha_tab_min_R[i]=new double[3];
	 double **alpha_tab_liberec = new double* [n_big_tria*2];
	 for (int i=0;i<2*n_big_tria;i++) alpha_tab_liberec[i]=new double[3];
	      
	 int domin=0;		// minimization of the eta_df_1 and eta_r estimator
	 int dojump=0;		// jump strategy
	 int domin_R=0;		// minimization of the eta_df_1 + eta_df_2 + eta_r estimator at once
	 int doliberec=0;	// eta_r = 0 locally
	 switch (strategy) // choose a strategy involving different estimators
	 {
	    case 0: domin_R=0;domin=0;dojump=1;doliberec=0; break; // only jump estimator
	    case 1: domin_R=0;domin=1;dojump=0;doliberec=0; break; // only minimization est.
	    case 2: domin_R=0;domin=1;dojump=1;doliberec=0; break; // minimum from both above
	    case 3: domin_R=1;domin=0;dojump=0;doliberec=0; break; // only minimization R.
	    case 4: domin_R=1;domin=1;dojump=1;doliberec=1; break; // do minimum from all possible strategies
	    case 5: domin_R=0;domin=0;dojump=0;doliberec=1; break; // special type by martin
	    case 6: domin_R=1;domin=0;dojump=1;doliberec=1; break; // min from all except min_1
	    default:
	    cout << "CEMRACS: unhandled strategy " << strategy << endl;
	  }
	      
	 // here, compute Poincare or Friedrichs constant based on the constants.pdf file : this constant is stored as C_F_or_C_P
	 double C_F_or_C_P = (boundary==0) ?
	        compute_C_P(2*n_big_tria, A_triangle_coor, vx, vy, boundary) :
	        compute_C_F(2*n_big_tria, A_triangle_coor, vx, vy, boundary);
			     
	// now compute th 
	 if (domin)   compute_th_minimization_1   (alpha_tab_min,   n_big_tria, A_triangle_coor, vx, vy, mp, sol, stack, boundary, C_F_or_C_P);
	 if (dojump)  compute_th_jumping        (alpha_tab_jump,  n_big_tria, A_triangle_coor, vx, vy, mp, sol, stack, boundary, C_F_or_C_P);
	 //	 if (dojump)  compute_th_minimization_2        (alpha_tab_jump,  n_big_tria, A_triangle_coor, vx, vy, mp, sol, stack, boundary, C_F_or_C_P);
	 if (domin_R) compute_th_minimization_2 (alpha_tab_min_R, n_big_tria, A_triangle_coor, vx, vy, mp, sol, stack, boundary, C_F_or_C_P);
	 if (doliberec) compute_th_liberec (alpha_tab_liberec, n_big_tria, A_triangle_coor, vx, vy, mp, sol, stack, boundary, C_F_or_C_P);

	// STEP 4 : computation of estimators based on th
	// using the computed values in alpha_tab, compute the estimators
	double est_DF=0,est_R=0,est_DF_1=0, est_DF_2=0;
	double est_DF_min=0,est_R_min=0,est_DF_1_min=0,est_DF_2_min=0;
	double est_DF_min_R=0,est_R_min_R=0,est_DF_1_min_R=0,est_DF_2_min_R=0;
	double est_DF_jump=0,est_R_jump=0,est_DF_1_jump=0,est_DF_2_jump=0;
	double est_DF_liberec=0,est_R_liberec=0,est_DF_1_liberec=0,est_DF_2_liberec=0;
	double value_min, value_jump, value_min_R, value_liberec;
	if (domin)   value_min   = compute_estimators(alpha_tab_min  ,C_F_or_C_P,A_triangle_coor,vx,vy,n_big_tria,mp,sol, stack,boundary,est_DF_min,est_R_min, est_DF_1_min, est_DF_2_min);
	if (dojump)  value_jump  = compute_estimators(alpha_tab_jump ,C_F_or_C_P,A_triangle_coor,vx,vy,n_big_tria,mp,sol, stack,boundary,est_DF_jump,est_R_jump, est_DF_1_jump, est_DF_2_jump);
	if (domin_R) value_min_R = compute_estimators(alpha_tab_min_R,C_F_or_C_P,A_triangle_coor,vx,vy,n_big_tria,mp,sol, stack,boundary,est_DF_min_R,est_R_min_R, est_DF_1_min_R, est_DF_2_min_R);
	if (doliberec) value_liberec = compute_estimators(alpha_tab_liberec,C_F_or_C_P,A_triangle_coor,vx,vy,n_big_tria,mp,sol, stack,boundary,est_DF_liberec,est_R_liberec, est_DF_1_liberec, est_DF_2_liberec);

	
	// choose which estimator to use
	double vv;
	switch (strategy)
	{
	    case 0: est_DF=est_DF_jump; 
		    est_R=est_R_jump; 
		    est_DF_1=est_DF_1_jump;
		    est_DF_2=est_DF_2_jump;
		    break;
    	    case 1: est_DF=est_DF_min; 
		    est_R=est_R_min; 
		    est_DF_1=est_DF_1_min;
		    est_DF_2=est_DF_2_min;
		    break;
	    case 2: if (value_min>value_jump)
		    { 
			est_DF=est_DF_jump; 
			est_R=est_R_jump; 
			est_DF_1=est_DF_1_jump;
			est_DF_2=est_DF_2_jump;
		    } else
		    { 
			est_DF=est_DF_min; 
			est_R=est_R_min; 
			est_DF_1=est_DF_1_min;
			est_DF_2=est_DF_2_min;
		    } 
		    break;
    	    case 3: est_DF=est_DF_min_R; 
		    est_R=est_R_min_R; 
    	    	    est_DF_1=est_DF_1_min_R;
		    est_DF_2=est_DF_2_min_R;
		    break;
	    case 4:  vv=value_jump;             
		    { 
			est_DF=est_DF_jump; 
			est_R=est_R_jump; 
			est_DF_1=est_DF_1_jump;
			est_DF_2=est_DF_2_jump;
		    }
		    if (value_min  <vv)  
		    { 
			vv=value_min; 
			est_DF=est_DF_min; 
			est_R=est_R_min; 
			est_DF_1=est_DF_1_min;
			est_DF_2=est_DF_2_min;
		    }
		    if (value_min_R<vv)  
		    {              
			est_DF=est_DF_min_R; 
			est_R=est_R_min_R; 
			est_DF_1=est_DF_1_min_R;
			est_DF_2=est_DF_2_min_R;
		    }
		    if (value_liberec<vv)  
		    {              
			est_DF=est_DF_liberec; 
			est_R=est_R_liberec; 
			est_DF_1=est_DF_1_liberec;
			est_DF_2=est_DF_2_liberec;
		    }
		    break;
	    case 5 : 	est_DF=est_DF_liberec; 
			est_R=est_R_liberec; 
			est_DF_1=est_DF_1_liberec;
		    	est_DF_2=est_DF_2_liberec;
			break;

	    case 6:  vv=value_jump;             
		    { 
			est_DF=est_DF_jump; 
			est_R=est_R_jump; 
			est_DF_1=est_DF_1_jump;
			est_DF_2=est_DF_2_jump;
		    }
		    if (value_min_R<vv)  
		    {              
			est_DF=est_DF_min_R; 
			est_R=est_R_min_R; 
			est_DF_1=est_DF_1_min_R;
			est_DF_2=est_DF_2_min_R;
		    }
		    if (value_liberec<vv)  
		    {              
			est_DF=est_DF_liberec; 
			est_R=est_R_liberec; 
			est_DF_1=est_DF_1_liberec;
			est_DF_2=est_DF_2_liberec;
		    }
		    break;
	    default: cout << "CEMRACS: unhandled strategy " << strategy << endl;
	}
			
	// fill the output fields		
	vector_eta_DF[global_vertex_number] = est_DF*est_DF;
	vector_eta_DF_1[global_vertex_number] = est_DF_1;
	vector_eta_DF_2[global_vertex_number] = est_DF_2;
	vector_eta_R[global_vertex_number] = est_R*est_R;
	// compute error on the dual volume:
	vector_err_v[global_vertex_number] =0;
	for (Int4 i=0 ; i<2*n_big_tria ; i++) vector_err_v[global_vertex_number] += qformula_exact_uni(vx,vy,A_triangle_coor[i][0],A_triangle_coor[i][1],A_triangle_coor[i][2],A_triangle_coor[i][3], mp, sol, stack);;

	// deletion of alpha_tabs
	for (int i=0 ; i<2*n_big_tria ; i++)
	{
	      delete [] alpha_tab_min[i];
	      delete [] alpha_tab_jump[i];
	      delete [] alpha_tab_min_R[i];
	      delete [] alpha_tab_liberec[i];
	}
	delete [] alpha_tab_min;
	delete [] alpha_tab_min_R;
	delete [] alpha_tab_jump;
	delete [] alpha_tab_liberec;

	// STEP 5: destroy all the A_triangle_coor generated in create_A_triangle_coor
	for (int i=0;i<n_big_tria*2;i++) delete [] A_triangle_coor[i];
	delete [] A_triangle_coor;

	  }
	}
    }
   // delete auxiliary stuff
   delete [] aux_array;

  // CEMRACS : now fill the output files with triangles information (quite simple) and then with data in the same time
  if (medit)
  {
    Int4 ni=0;
    for (Int4  it = 0; it < g_nt; it++) for (Int4 iv=0;iv<3;iv++)
    {
      int giv=Th.Number(Th(it)[iv]);
      of_mesh_eta << (sqrt(vector_eta_DF[giv])+sqrt(vector_eta_R[giv])) << endl;
      of_mesh_eta_DF << sqrt(vector_eta_DF[giv]) << endl;
      of_mesh_eta_R << sqrt(vector_eta_R[giv]) << endl;
      of_mesh_err << sqrt(vector_err_v[giv]) << endl;
      of_mesh_exact << ex(Th(it)[iv].r.x,Th(it)[iv].r.y,mp,sol,stack) << endl;
    }
    // mesh generation : vertices
    for (Int4  it = 0; it < g_nt; it++) for (Int4 iv=0;iv<3;iv++) of_mesh << Th(it)[iv].r.x << " " << Th(it)[iv].r.y << " 0 " << endl;      
    of_mesh << "Triangles " << g_nt << endl;
    ni=0;
     for (Int4  it = 0; it < g_nt; it++) { for (Int4 iv=0;iv<3;iv++) { ni++; of_mesh << ni << " ";  } of_mesh << "0" << endl; }
    of_mesh << "End" << endl;
    // copy the mesh files because medit is STUPID PIECE OF ,,., code
    system(("cp "+*fn+"_mesh.mesh "+*fn+"_mesh_err.mesh").c_str());
    system(("cp "+*fn+"_mesh.mesh "+*fn+"_mesh_eta.mesh").c_str());
    system(("cp "+*fn+"_mesh.mesh "+*fn+"_mesh_eta_DF.mesh").c_str());
    system(("cp "+*fn+"_mesh.mesh "+*fn+"_mesh_eta_R.mesh").c_str());
    system(("cp "+*fn+"_mesh.mesh "+*fn+"_mesh_exact.mesh").c_str());
    of_mesh.close();     of_mesh_eta.close();    of_mesh_eta_DF.close();    of_mesh_eta_R.close();    of_mesh_err.close();    of_mesh_exact.close();
  } // medit ends here
  // export
  double estimator=0, etaDF, etaR;
  for (Int4  iv = 0; iv < g_nv; iv++) 
    {
      etaDF = sqrt(vector_eta_DF[iv]);
      etaR = sqrt(vector_eta_R[iv]);
      vector_vertices[iv] = ((etaDF+etaR)*(etaDF+etaR));  //  store (eta_DF_D + et_R_D)^2 for each vertex 
    }
  return SetAny<double>(estimator);
}  



//   to add the name myfunction to the freefem++ table 
class Init { public:
  Init();
};
Init init;
Init::Init(){
  Global.Add("AposterioCemracs2007","(",new OneOperatorCode<AposterioCemracs2007>);
}

