double PI=bamg::Pi;

void intersection(double a1, double a2, double b1, double b2, double c1, double c2, double d1, double d2, double &x1, double &x2)
{
  double t2nom   = (c2-a2)*(b1-a1)-(b2-a2)*(c1-a1);
  double t2denom = (c2-d2)*(b1-a1)-(b2-a2)*(c1-d1);
  double t2 = t2nom/t2denom;
  x1 = c1+t2*(d1-c1);
  x2 = c2+t2*(d2-c2);
}

double distance(double a1, double a2, double b1, double b2)
{
  return sqrt((b1-a1)*(b1-a1)+(b2-a2)*(b2-a2));
}

void barycenter(double &x1, double &x2, double a1, double a2, double b1, double b2, double c1, double c2) // c1,c2 ... center
{
  x1 = (a1+b1+c1)/3.0;
  x2 = (a2+b2+c2)/3.0;
}

double angle(double a1, double a2, double b1, double b2, double c1, double c2) // c1,c2 ... center
{
  double v1 = a1-c1;
  double v2 = a2-c2;
  double u1 = b1-c1;
  double u2 = b2-c2;
  double v = distance(a1,a2,c1,c2);
  double u = distance(b1,b2,c1,c2);
  double product = (v1*u1+v2*u2)/u/v;
  return acos(product);
}

double vectorangle(double u1, double u2)
{
//  double pi = bamg::Pi;
//  cout << "u1 = "<< u1 << " u2 = "<< u2<<endl;
  if (u2 == 0) return (u1 <0) ? PI : 0;
  if (u1 == 0) return (u2 <0) ? 3.0/2.0*PI : PI/2.0;
  double basis = atan(fabs(u2/u1)); // first quadrant
//  cout <<" basis = " << basis << endl;
  if (u1 >= 0 && u2 > 0) return basis;
  if (u1 < 0 && u2 > 0) return PI-basis;
  if (u1 < 0 && u2 < 0) return PI+basis;
  if (u1 >=0 && u2 < 0) return 2.0*PI-basis;
  cout << "vectorangle unhandled case "<< endl;
  return 0;
}

int compute_angle_interval(double &angle_d, double &angle_u, double c1, double c2, double a1, double a2, double b1, double b2)
{
  if (a1==c1 && a2==c2) return 2; // don't take this one into account
  if (b1==c1 && b2==c2) return 2;
  double anga = vectorangle(a1-c1,a2-c2); // in [0,2pi)
  double angb = vectorangle(b1-c1,b2-c2); // in [0,2pi)
//  cout << "angle1="<<anga<<" angle2="<<angb<<" c1="<<c1<<" c2="<<c2<<" a1="<<a1 <<" a2="<<a2<<" b1="<<b1<<" b2="<< b2<<	endl;
  // now : the angle between three points is always between 0 and pi !
  if (fabs(anga-angb)>PI) // this is the case, where we go below zero (mod 2pi)
  {
     angle_d = (anga > angb) ? anga : angb;    
     angle_u = (anga > angb) ? angb : anga;
  } else
  {
     angle_d = (anga < angb) ? anga : angb;    
     angle_u = (anga < angb) ? angb : anga;
  }
  if (anga==angb) return 0; // no interval here
  return 1;
}

int isin(double u, double a1, double a2)
// the intersection is made as a closed intervals.
{
  if (u<0 || u > 2.0*PI) { cout << "isin out of range " << u << endl; return 0; }
  if (a1 > a2)
  {
    if (u >= a1 || u <= a2) return 1;
  } else
  {
     if (u >= a1 && u <= a2) return 1;
  }
  return 0;
}

int interval_intersection(double &x1,double &x2, double a1, double a2,double b1,double b2)
// warning : this intersection function works only if the maximal length of each interval does not exceed Pi
{
//  if (a1==a2 || b1==b2) { cout << "void interval."<<endl; x1=0;x2=0;return 0; }
  x1=0;x2=0;
  // x1 section
  if (a1==b1) x1=a1; else
  if (isin(a1,b1,b2)) x1=a1;else
  if (isin(b1,a1,a2)) x1=b1;else
  return 0; // no intersection
  // x2 section
  if (a2==b2) x2=a2; else
  if (isin(a2,b1,b2)) x2=a2;else
  if (isin(b2,a1,a2)) x2=b2;else
  return 0;
  return 1;
}


int is_intersection_nonempty(int ne, int n,int k,double **T_d,double **T_u,int **T_is,double lower, double upper)
// recursion, hopefully it will not be too deeep !!!
{
  if (k==n) return 1; // we found non-empty intersection 
  double x1,x2; int is;
  for (int i=0;i<3*ne;i++)
  {
    if (T_is[k][i])
    {
      is = interval_intersection(x1,x2,T_d[k][i],T_u[k][i],lower,upper);
      if (is==1) // we found non-empty intersection, let's go recursive to reach the level 'k==n'
      {
        is = is_intersection_nonempty(ne,n,k+1,T_d,T_u,T_is,x1,x2);
        if (is==1) return 1;  
      }
    }
  }
  return 0;
}


int iscondition27(int n, double **A_triangle_coor, double v1, double v2, int ne, double **A_edges)
// n ... number of triangles around [v1,v2]
// A_triangle_coor ... coordinates of all vertices of all triangles around [v1,v2] in the anticlockwise direction 
// ne ... number of boundary edges
// A_edges ... coordinates of all boundary edges of the dual volume A_edges[i] = { a_x,a_y,b_x,b_y }
{
  // init
  double **T_d = new double* [n];
  double **T_u = new double* [n];
  int **T_is = new int*[n];
  int ret[3];
  double angle_d[3],angle_u[3];
  for (int i=0;i<n;i++) T_d[i]=new double[3*ne];
  for (int i=0;i<n;i++) T_u[i]=new double[3*ne];
  for (int i=0;i<n;i++) T_is[i]=new int[3*ne];
    
  for (int it=0;it<n;it++)  // go through all triangles in the field
  {
    for (int ie=0;ie<ne;ie++) // go through all boudnary edges
    {
      ret[0] = compute_angle_interval(angle_d[0],angle_u[0],v1,v2,A_edges[ie][0],A_edges[ie][1],A_edges[ie][2],A_edges[ie][3]);
      ret[1] = compute_angle_interval(angle_d[1],angle_u[1],A_triangle_coor[it][0],A_triangle_coor[it][1],A_edges[ie][0],A_edges[ie][1],A_edges[ie][2],A_edges[ie][3]);
      ret[2] = compute_angle_interval(angle_d[2],angle_u[2],A_triangle_coor[it][2],A_triangle_coor[it][3],A_edges[ie][0],A_edges[ie][1],A_edges[ie][2],A_edges[ie][3]);
     for (int j=0;j<3;j++)
     {
         T_is[it][ie*3+j] = (ret[j]==1) ? 1 : 0;
	 T_d[it][ie*3+j] = angle_d[j];
	 T_u[it][ie*3+j] = angle_u[j];
     }

     } // end for it
  } // end for ie
  
  // now the important part = do the intersection of all intervals
  int is_nonempty=0; // it is empty
  for (int ie=0;ie<3*ne && is_nonempty==0;ie++)
  {    
    if (T_is[0][ie]==1)
    {
      is_nonempty=is_intersection_nonempty(ne,n,0,T_d,T_u,T_is, T_d[0][ie], T_u[0][ie]);
    }
  }
  // delete
  for (int i=0;i<n;i++) delete [] T_d[i];
  for (int i=0;i<n;i++) delete [] T_u[i];
  for (int i=0;i<n;i++) delete [] T_is[i];
  delete [] T_u;
  delete [] T_d;
  delete [] T_is;
  return is_nonempty; 
}


int isconvex(int nt, double **A_triangle_coor, double v1, double v2,
		    int boundary)
// we have a set of triangles around vertex v1,v2 - closed or not.
{
  if (nt==1) return 1; // only one triangle = convex. ok.
//  C_F = 0;
//  double C_P = 0;
    int convex=1;
    //  check for convexity
    // 1. test the convexity of the domain. assuming the ordering of the A_triangle_coor
    for (int i=(boundary==0)?0:1;i<nt && convex==1;i++) // 2*i = first triangle in big triangle, 2*i+1 = second triangle in the big triangle, they share vertex [vx,vy] and barycenter [TAB[2],TAB[3]]
    {
      double a1, a2, b1, b2;
      if (i==0) // the boundary case, means that the convexity have to be check with the (n_big_tria-1) big triangle
      {
        a1=A_triangle_coor[0][2];
        a2=A_triangle_coor[0][3];
	b1=A_triangle_coor[nt-1][0];
        b2=A_triangle_coor[nt-1][1];
      } else
      {
        a1=A_triangle_coor[i][2];
        a2=A_triangle_coor[i][3];
	b1=A_triangle_coor[i-1][0];
        b2=A_triangle_coor[i-1][1];
      }
      // constuct the midpoints of the common edge between i and i-1 triangle, resp. 0 and n_big_tria-1
      double c1 = A_triangle_coor[i][0];
      double c2 = A_triangle_coor[i][1];
      // compute the intersection point x1,x2
      double x1,x2;
      intersection(a1,a2,b1,b2,c1,c2,v1,v2,x1,x2);
      double dist1 = distance(x1,x2,v1,v2);
      double dist2 = distance(c1,c2,v1,v2);
      if (dist1-dist2 > 1e-10)  convex=0;   
    }
    if (convex==1)
    {
      if (boundary!=0) 
      {
        // SPECIAL CODE : compare the angles to find out, if the non-internal dual volume is convex or not due to boundary shaping
	// a. compute the angle between boundary points
	double angle1 = angle(A_triangle_coor[0][0],A_triangle_coor[0][1],A_triangle_coor[nt-1][2],A_triangle_coor[nt-1][3],v1,v2);
	// b. compute the angle between inner parts of the dual volume : there is at least 2 triangles, i.e., n_big_tria>0
	double angle2 = angle(A_triangle_coor[0][0],A_triangle_coor[0][1],A_triangle_coor[nt-1][0],A_triangle_coor[nt-1][1],v1,v2);
	double angle3 = angle(A_triangle_coor[nt-1][0],A_triangle_coor[nt-1][1],A_triangle_coor[nt-1][2],A_triangle_coor[nt-1][3],v1,v2);
	// c. now do the sum:
	if (angle1 - (angle2+angle3)> -1e-10) convex = 1; else convex = 0;
//      cout << "CEMRACS : domain convex= "<<convex<< "  and boundary "<< boundary << endl;
      }
    }
 return convex;
}



double compute_C_P(int n, double **A_triangle_coor, double v1, double v2,int boundary)
{
  int convex = isconvex(n, A_triangle_coor,v1,v2,boundary);
  double C_P = 0;
  if (convex==1)
  {
      C_P = 1.0/PI/PI;
  } else
  {
       double C2D2=0;
       // now go through all small triangles and compute the value of gammaK
       for (int i=0;i<n;i++)
       {
         double a1 = A_triangle_coor[i][0];
         double a2 = A_triangle_coor[i][1];
         double b1 = A_triangle_coor[i][2];
         double b2 = A_triangle_coor[i][3];
	 double auxp1 = v1 + (b2-a2); // aux point on the normal that goes through the vertex v1,v2
	 double auxp2 = v2 - (b1-a1);
	 double x1,x2; // the intersection 
	 intersection(a1,a2,b1,b2,auxp1,auxp2,v1,v2,x1,x2);
	 double distHEIGHT = distance(v1,v2,x1,x2); if (distHEIGHT<=1e-30) cout << "CEMRACS warning : small height. "<<distHEIGHT << endl;
	 double distALPHA = distance(a1,a2,x1,x2);
	 double distBETA = distance(b1,b2,x1,x2);
	 double tanALPHA = distALPHA/distHEIGHT;
	 double tanBETA = distBETA/distHEIGHT;
	 double gammaK = (tanALPHA > tanBETA) ? tanALPHA : tanBETA; // maximum
	 double C2K2=2.0+2.0*gammaK*(sqrt(1.0+gammaK*gammaK)+gammaK);
	 if (C2K2 > C2D2) C2D2=C2K2; // take maximum if reached
       }
       // final step : compute the Poincare constant
        C_P =  C2D2/PI/PI;
//	cout << "compute_C_P ["<<v1<<","<<v2<<"] non-convex : C_P = "<<C_P<<endl;
  } 

  return C_P;
}

double compute_diameter(int n, double **A, double v1, double v2)
// finds the maximal edge in the dual volume/ triangles and takes it twice as a diameter
{
  double max_h=0;
  for (int i=0;i<n;i++)
  {
         double a1 = A[i][0];
         double a2 = A[i][1];
         double b1 = A[i][2];
         double b2 = A[i][3];
	 // now copmute h_D
	 double h1 = distance(a1,a2,v1,v2);
	 double h2 = distance(b1,b2,v1,v2);
	 double h = (h1>h2)?h1:h2;
	 if (h > max_h) max_h=h;
  }
  return 2.0*max_h;
}

double compute_minimal_diameter(int n, double **A, double v1, double v2)
// finds the maximal edge in the dual volume/ triangles and takes it twice as a diameter
{
  double min_h=1e10;
  for (int i=0;i<n;i++)
  {
         double a1 = A[i][0];
         double a2 = A[i][1];
         double b1 = A[i][2];
         double b2 = A[i][3];
	 // now copmute h_D
	 double h1 = distance(a1,a2,v1,v2);
	 double h2 = distance(b1,b2,v1,v2);
	 double h3 = distance(b1,b2,a1,a2);
	 double h = h1;
	 if (h < h2) h=h2;
	 if (h < h3) h=h3;
	 // h ... is the diameter
	 if (h < min_h) min_h=h;
  }
  return min_h; //return the minimum
}

double area_K(double x1, double y1, double x2, double y2, double x3, double y3)
{
  return 0.5*fabs(x1*y3-x1*y2+x2*y1-x2*y3+x3*y2-x3*y1);
}

double compute_max_edge_triangle(double x1, double y1, double x2, double y2, double x3, double y3)
{
  double h1 = distance(x2,y2,x3,y3);
  double h2 = distance(x1,y1,x3,y3);
  double h3 = distance(x1,y1,x2,y2);
  double h = h1;
  if (h < h2) h = h2;
  if (h < h3) h = h3;
  return h;
}

double compute_min_edge_triangle(double x1, double y1, double x2, double y2, double x3, double y3)
{
  double h1 = distance(x2,y2,x3,y3);
  double h2 = distance(x1,y1,x3,y3);
  double h3 = distance(x1,y1,x2,y2);
  double h = h1;
  if (h > h2) h = h2;
  if (h > h3) h = h3;
  return h;
}


double compute_C_F(int n, double **A, double v1, double v2,int boundary)
{
  if (boundary==0) return 1.0; // don't know what to do
  double C_F;
  int ne=2;
  double **A_edges = new double* [ne];
  for (int i=0;i<ne;i++) A_edges[i]=new double[4];
  // create the boundary edges field (only 2 edges in the dual volume case)
  A_edges[0][0]=v1; 
  A_edges[0][1]=v2; 
  A_edges[0][2]=A[0][0]; 
  A_edges[0][3]=A[0][1];
  // edge 1
  A_edges[1][0]=v1; 
  A_edges[1][1]=v2; 
  A_edges[1][2]=A[n-1][2]; 
  A_edges[1][3]=A[n-1][3];
  // check the condition 2.7
  
  int is27 = iscondition27(n,A,v1,v2,ne,A_edges);

  if (is27) C_F=1.0;
  else
  {
    double delta_D = 0;
    double area_D = 0;
    double max_lenght_edge=0;
    for (int i=0;i<n;i++)
    {
         double a1 = A[i][0];
         double a2 = A[i][1];
         double b1 = A[i][2];
         double b2 = A[i][3];
	 double auxp1 = v1 + (b2-a2); // aux point on the normal that goes through the vertex v1,v2
	 double auxp2 = v2 - (b1-a1);
	 double x1,x2; // the intersection 
	 intersection(a1,a2,b1,b2,auxp1,auxp2,v1,v2,x1,x2);
	 double distHEIGHT = distance(v1,v2,x1,x2); if (distHEIGHT<=1e-30) cout << "CEMRACS warning : small height. "<<distHEIGHT << endl;
	 double distALPHA = distance(a1,a2,x1,x2);
	 double distBETA = distance(b1,b2,x1,x2);
	 double tanALPHA = distALPHA/distHEIGHT;
	 double tanBETA = distBETA/distHEIGHT;
	 double delta_K = tanALPHA+tanBETA;
	 // now put delta_D add
	 delta_D+=delta_K;
	 area_D+=area_K(v1,v2,a1,a2,b1,b2);
	 // now copmute h_D
    }
    //evaluate
    double h_D = compute_diameter(n,A,v1,v2);
    C_F = 2.0*area_D*delta_D/h_D/h_D;
  }
  // delete  
  for (int i=0;i<ne;i++) delete [] A_edges[i];
  delete [] A_edges;
  return C_F;
}



