#include  <iostream>
#include  <cfloat>
using namespace std;
#include "error.hpp"
#include "AFunction.hpp"
#include "rgraph.hpp"
#include "RNM.hpp"
#include "fem.hpp"
#include "FESpace.hpp" 
#include "MeshPoint.hpp"
#include "BamgFreeFem.hpp"


#include "Operator.hpp"

#include <set>
#include <vector>
#include <fstream>

#include "lex.hpp"
#include "Mesh2.h"

#include "BamgFreeFem.hpp"
#include "lgfem.hpp"

// auxiliary functions
#include "cemracs2007aux.cpp"

using namespace Fem2D;

/**********************************/
/* good functions from the script*/
double rhs(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[3])(stack));
}

double rhs_reaction(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{ // return f-rp_h
  mp->set(x,y,0);
  return GetAny<double>( (*sol[3])(stack))-GetAny<double>( (*sol[7])(stack))*GetAny<double>( (*sol[0])(stack));
}

double reaction(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[7])(stack));
}

double ex(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[4])(stack));
}
double ex_dx(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[5])(stack));
}
double ex_dy(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[6])(stack));
}

double ph_dx(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[1])(stack));
}

double ph_dy(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[2])(stack));
}
/* end of good functios           */
/**********************************/



/* This code is essential : computes  the estimator */
class AposterioCemracs2007 :   public E_F0mps { public:
  typedef double  Result;
  
  static basicAC_F0::name_and_type name_param[] ;
  static const int n_name_param =1;
  static const int n_output_param = 2;
  
  int nbsol;    
  Expression nargs[n_name_param];
  Expression getmesh;
  Expression em_output[n_output_param]; 
  int  typesol[100];
  vector<Expression> sol;
  
  double arg(int i,Stack stack,double a) const { return nargs[i] ? GetAny<double>( (*nargs[i])(stack) ): a;}
  long arg(int i,Stack stack,long a) const{ return nargs[i] ? GetAny<long>( (*nargs[i])(stack) ): a;}
  bool arg(int i,Stack stack,bool a) const{ return nargs[i] ? GetAny<bool>( (*nargs[i])(stack) ): a;}
  int arg(int i,Stack stack,int a) const{ return nargs[i] ? GetAny<int>( (*nargs[i])(stack) ): a;}
  
  AposterioCemracs2007(const basicAC_F0 & args) :nbsol(args.size()-1),sol(args.size()-1)
  {   
      args.SetNameParam(n_name_param,name_param,nargs);
      getmesh=to<pmesh>(args[0]); 
     const E_Array * expoutput = dynamic_cast<const E_Array *>(nargs[0]);
     if(expoutput)
     {
       if(expoutput->nbitem()!=n_output_param) ExecError("\nSorry we wait an array with 7 components in: output=[vector_triangles(P0) vector_vertices(P1) vector_eta_DF(P1) vector_eta_R(P1)  vector_err_k(P0) vector_err_v(P1) hd(P1) ]");
       for (int i=0;i<n_output_param;i++) em_output[i] = (*expoutput)[i];
     } else ExecError("\n CEMRACS_exec:  output=[] is missing ! ");
   }  



    static ArrayOfaType  typeargs() { return  ArrayOfaType(atype<pmesh>(),true);}
    static  E_F0 * f(const basicAC_F0 & args){ return new AposterioCemracs2007(args);} 
    AnyType operator()(Stack s) const ;    
    operator aType () const { return atype<pmesh>();}         
};

 basicAC_F0::name_and_type AposterioCemracs2007::name_param[AposterioCemracs2007::n_name_param] = {
       {   "output"          ,  &typeid(E_Array)}  // 1 arrays 
    };



AnyType AposterioCemracs2007::operator()(Stack stack) const 
{
// DECLARATION OF VARIABLES
  using namespace bamg;
  double pi = bamg::Pi;
  double cutoffradian          = -1.0* pi/180. ;
  double *mm_output[n_output_param]; for (int i=0;i<n_output_param;i++) mm_output[i]=0;
  using Fem2D::MeshPoint;
  using Fem2D::Mesh;
  Mesh * Thh = GetAny<pmesh>((*getmesh)(stack));
  ffassert(Thh);
  Triangles * oTh = msh2bamg(*Thh,cutoffradian);
  Triangles &Th(*oTh);

  Th.ReMakeTriangleContainingTheVertex();   // TODO : what does it do ? now idea...
  MeshPoint* mp(Fem2D::MeshPointStack(stack)); // ok - MeshPoint stack is reffered in mp
  long g_nt = Thh->nt; // the number of triangles
  long g_nv = Thh->nv; // the number of vertices

  for (int i=0;i<n_output_param;i++) mm_output[i]= GetAny<double *>( (*em_output[i])(stack) );
  double &c1=*mm_output[0];
  double &c2=*mm_output[1];

   // CEMRACS : try to go through vertices and find all adjacent triangles
   int *aux_array = new int[g_nv]; // this is only an indicator used to be sure that we go through all vertices only once
   double *c1_torso_array = new double[g_nv]; 
   for (Int4 iv =0;iv<g_nv;iv++) aux_array[iv]=0;
    // go through all triangles and do algorithms for all vertices
    for (Int4  it = 0; it < g_nt; it++) 
    {
        bamg::Triangle & Kinit = Th(it); // current triangle we work on
	for (Int4 iv = 0; iv < 3; iv++)
	{
	  Int4 global_vertex_number = Th.Number(Kinit[iv]); // vertex we are messing around
    	  if (aux_array[global_vertex_number]==0)
	  {
	    // STEP -1 : initialization
  	    double vx = Kinit[iv].r.x;
	    double vy = Kinit[iv].r.y;
	    Int4 numKinit = it;               // initial number of the triangle we started from
	    Int4 numKcur =  numKinit;         // current number of the triangle we are in
	    Int4 done =0;                     // is the while loop over ?
	    Int4 boundary = 0; // are we on the boundary ?
	    Int4 numKstart = -1;
	  
	    Int4 num_adj_edge = 0; // number of the edge the next triangle is adjacent to 
	    Int4 local_vertex_num = iv;
	    Int4 n_big_tria = 0;   // number of big triangles around the vertex. 0 = there is no adjacent triangles
	    // STEP 0: now put information that this vertex was already computed
	    aux_array[global_vertex_number] = 1;
	    // STEP 1a: now go through the triangles  clock wise to find the boundary or simply count the global number of big triangles
	    while (done==0)
	    {
	      // now find the local index of the vertex in the numKcur triangle
//              bamg::Triangle & K = Th(numKcur); // current triangle we work on
	      n_big_tria++;
	      local_vertex_num=-1;
	      for (int i=0;i<3;i++) if (Th.Number(Th(numKcur)[i])==global_vertex_number) local_vertex_num = i;
	      if (local_vertex_num == -1) cout << "CEMRACS error : no match in the local vertex number : the vertex is not in the triangle !!! panic ..." << endl;
	      num_adj_edge = (local_vertex_num+2)%3;  // number of the edge for the adjacent triangle to go : clockwse
	      if (Th.Number(*Th(numKcur).Adj(num_adj_edge)) >= g_nt)  // we reached the edge from the left clock wise direction !!!
	      {
	         boundary = 1;
		 done = 1;
		 numKstart = numKcur;
	      } 
	      numKcur = Th.Number(*Th(numKcur).Adj(num_adj_edge));
	      if (numKcur == numKinit) // hooray, we meet the initial triangle again
	      {
	        done = 1;
		boundary = 0;
		numKstart = numKinit;
	      }
	    }
//	    cout << "vertex "<< global_vertex_number <<" number of triangles around " << n_big_tria <<" boundary "<< boundary<< endl;
	    // STEP 1b : now adjust the number of triangles in the case of boundary : go anti-clock wise from initial triangle and only count the rest of the triangles
	    if (boundary==1)
	    {
	    n_big_tria--;
	    numKcur=numKinit;
	    boundary=0;
	    done=0;
	    while (done==0)
  	      {
	      // now find the local index of the vertex in the numKcur triangle
//              bamg::Triangle & K = Th(numKcur); // current triangle we work on
	      n_big_tria++;
	      local_vertex_num=-1;
	      for (int i=0;i<3;i++) if (Th.Number(Th(numKcur)[i])==global_vertex_number) local_vertex_num = i;
	      if (local_vertex_num == -1) cout << "CEMRACS error : no match in the local vertex number : the vertex is not in the triangle !!! panic ..." << endl;
	      num_adj_edge = (local_vertex_num+1)%3;  // number of the edge for the adjacent triangle to go : anticlockwise
	      if (Th.Number(*Th(numKcur).Adj(num_adj_edge)) >= g_nt)  // we reached the edge from the left clock wise direction !!!
	      {
	         boundary = 1;
		 done = 1;
	      } 
	      numKcur = Th.Number(*Th(numKcur).Adj(num_adj_edge));
	      if (numKcur == numKinit) // hooray, we meet the initial triangle again
	      {
	        done = 1;
		boundary = 0;
	      }
	      }
  	    if (boundary==0) cout << "CEMRACS critical error : no boundary detected in the 2nd run !!!" << endl;
//	    else cout << "CEMRACS : number of triangles around boundary vertex is " << n_big_tria << endl;
	    }
	    
	    // STEP 2 : create the desired field of triangles 
	    // now go through all triangles from left to right and put data into the fileds
	    double **A_triangle_coor = new double*[n_big_tria]; // field of small triangle vertex coordinates
	    Int4 *A_triangle_indexes = new Int4[n_big_tria];    // field of number of big triangle the small is inside
	    for (int i=0;i<n_big_tria;i++) A_triangle_coor[i] = new double[4];  
	    done = 0;
	    int n_small_tria=0, n_tria_check=0;
	    numKcur=numKstart; // just start the loop from the most left positioned triangle (if boundary), go anticlockwise
	    while (done==0)
	    {
	      n_tria_check++;
	      if (n_tria_check > n_big_tria) cout << "CEMRACS error : number of triangles exceeded !!! " << endl;
	      // now find the local index of the vertex in the numKcur triangle
	      local_vertex_num=-1;
	      for (int i=0;i<3;i++) if (Th.Number(Th(numKcur)[i])==global_vertex_number) local_vertex_num = i;
	      if (local_vertex_num == -1) cout << "CEMRACS error : no match in the local vertex number : the vertex is not in the triangle !!! panic ..." << endl;
	      // a0. now set x1,y1, x2,x2, x0,y0
	         double x0 = Th(numKcur)[local_vertex_num].r.x;
		 double y0 = Th(numKcur)[local_vertex_num].r.y;
		 if (x0!=vx || y0!=vy) cout << "CEMRACS error : the vertex coordinates does not match local_vertex_num "<<endl;
	         double x1 = Th(numKcur)[(local_vertex_num+1)%3].r.x;
		 double y1 = Th(numKcur)[(local_vertex_num+1)%3].r.y;
	         double x2 = Th(numKcur)[(local_vertex_num+2)%3].r.x;
		 double y2 = Th(numKcur)[(local_vertex_num+2)%3].r.y;
	         
		 A_triangle_coor[n_small_tria][0] = x1;
		 A_triangle_coor[n_small_tria][1] = y1;
		 A_triangle_coor[n_small_tria][2] = x2;
		 A_triangle_coor[n_small_tria][3] = y2;

		 A_triangle_indexes[n_small_tria] = numKcur;
		 n_small_tria++;

	      // now go to the next triangle
	      if (n_tria_check==n_big_tria) done =1;
//	      cout << "n_tria_check "<<n_tria_check << " n_big_tria "<< n_big_tria << endl;
	      if (done==0)
	      {
	      num_adj_edge = (local_vertex_num+1)%3;  // number of the edge for the adjacent triangle to go : anticlockwse
	      if (Th.Number(*Th(numKcur).Adj(num_adj_edge)) >= g_nt)  // we reached the edge from the left clock wise direction !!!
		 cout << "CEMRACS critical error : reached boundary but was not supposed to !!!!" << endl;
	      numKcur = Th.Number(*Th(numKcur).Adj(num_adj_edge));
	      if (numKcur == numKstart) // hooray, we meet the initial triangle again
		 cout << "CEMRACS critical error : reached the initial triangle but was not supposed to !!!!" << endl;
	      }
	    }
	    
	      // STEP 3 : now, compute c1_torso  from the Carstensen and Funken paPER
	      double h_omega_v = compute_diameter(n_big_tria, A_triangle_coor, vx, vy);
	      double c_omega =(boundary==0) ? 
	          sqrt(compute_C_P(n_big_tria, A_triangle_coor, vx, vy, boundary))*h_omega_v :
	          sqrt(compute_C_F(n_big_tria, A_triangle_coor, vx, vy, boundary))*h_omega_v;
		  
 	      double h_T_min = compute_minimal_diameter(n_big_tria, A_triangle_coor, vx, vy);
	      double c1_local_torso = c_omega*c_omega/h_T_min/h_T_min;
	      	  
	    // STEP 4 : store the c1_local_torso into a field
	    c1_torso_array[global_vertex_number] = c1_local_torso;
		 
	    // STEP 5: destroy all the A_triangle_coor and A_triangle_indexes
	    for (int i=0;i<n_big_tria;i++) delete [] A_triangle_coor[i];
	    delete [] A_triangle_coor;
	    delete [] A_triangle_indexes;
	  }
	}
    }
   // delete stuff
   delete [] aux_array;
    
   // compute c1 here
   c1=0;
   for (int it=0;it<g_nt;it++)
   {  
        bamg::Triangle & K = Th(it); // current triangle we work on
	double Ksum = sqrt(c1_torso_array[Th.Number(K[0])]+
		           c1_torso_array[Th.Number(K[1])]+ 
		           c1_torso_array[Th.Number(K[2])]);
	if (Ksum > c1) c1 = Ksum; // take max over all triangles
   }
    
   // compute c2 here :  go through all triangles 
   double c2_torso_1 = 0; 
   double c2_torso_2 = 0; 
   for (int it=0;it<g_nt;it++)
   {
        bamg::Triangle & K = Th(it); // current triangle we work on
	double x0 = K[0].r.x;
	double y0 = K[0].r.y;
	double x1 = K[1].r.x;
	double y1 = K[1].r.y;
	double x2 = K[2].r.x;
	double y2 = K[2].r.y;
	double h_T = compute_max_edge_triangle(x0,y0,x1,y1,x2,y2);
	double area_T = area_K(x0,y0,x1,y1,x2,y2);
	double h_E_min = compute_min_edge_triangle(x0,y0,x1,y1,x2,y2);
	double torso_1 = 3.0*h_T*h_T*h_T/h_E_min/area_T;
//	double torso_2 = torso_1*1.0/2.0/sqrt(3.0)*(3.0+ h_T*h_T/area_T);
	double torso_2 = torso_1*1.0/2.0*sqrt(3.0)*(3.0+ h_T*h_T/area_T);
	// now store the maximal value
	if (torso_1 > c2_torso_1) c2_torso_1 = torso_1;
	if (torso_2 > c2_torso_2) c2_torso_2 = torso_2;
   }
   c2 = sqrt(c1*c2_torso_1 + c1*c1*c2_torso_2);
  
   //delete 
   delete [] c1_torso_array;
   double estimator = 0;
  return SetAny<double>(estimator);
}  



//   to add the name myfunction to the freefem++ table 
class Init { public:
  Init();
};
Init init;
Init::Init(){
  Global.Add("AposterioCemracs2007Classical","(",new OneOperatorCode<AposterioCemracs2007>);
}

