#include  <iostream>
#include  <cfloat>
using namespace std;
#include "error.hpp"
#include "AFunction.hpp"
#include "rgraph.hpp"
#include "RNM.hpp"
#include "fem.hpp"
#include "FESpace.hpp" 
#include "MeshPoint.hpp"
#include "BamgFreeFem.hpp"


#include "Operator.hpp"

#include <set>
#include <vector>
#include <fstream>

#include "lex.hpp"
#include "Mesh2.h"

#include "BamgFreeFem.hpp"
#include "lgfem.hpp"

// auxiliary functions
//#include "cemracs2007aux.cpp"
#include "EZ/EF.cpp"


using namespace Fem2D;

/**********************************/
/* good functions from the script*/

double rhs_reaction(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{ // return f-rp_h
  mp->set(x,y,0);
  return GetAny<double>( (*sol[3])(stack))-GetAny<double>( (*sol[7])(stack))*GetAny<double>( (*sol[0])(stack));
}

double reaction(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[7])(stack));
}

double ex(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[4])(stack));
}
double ex_dx(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[5])(stack));
}
double ex_dy(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[6])(stack));
}

double ph_dx(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[1])(stack));
}

double ph_dy(double x, double y, MeshPoint* mp, vector<Expression> sol, Stack &stack) 
{
  mp->set(x,y,0);
  return GetAny<double>( (*sol[2])(stack));
}
/* end of good functios           */
/**********************************/


// TABOR-20071007
/* This code is essential : computes  the estimator */
class Cemracs2007solver :   public E_F0mps { public:
  typedef double  Result;
  
  static basicAC_F0::name_and_type name_param[] ;
  static const int n_name_param =1;
  static const int n_output_param = 1;
  
  int nbsol;    
  Expression nargs[n_name_param];
  Expression getmesh;
  Expression em_output[n_output_param]; 
  // em_output[0] : eta_t ... eta_k
  // em_output[1] : eta_v
  // em_output[2] : eta_v_DF
  // em_output[3] : eta_v_R
  // em_output[4] : err_t ... err_k   = exact error per triangle
  // em_output[5] : err_v             = exact error per vertice = dual volume !
  // em_output[6] : h_d             = per vertex
  // em_output[7] : eta_v_DF_1      = per vertex
  // em_output[8] : eta_v_DF_2      = per vertex
  int  typesol[100];
  vector<Expression> sol;
  int nbcperiodic;
  Expression *periodic;
  
  
  double arg(int i,Stack stack,double a) const { return nargs[i] ? GetAny<double>( (*nargs[i])(stack) ): a;}
  long arg(int i,Stack stack,long a) const{ return nargs[i] ? GetAny<long>( (*nargs[i])(stack) ): a;}
  bool arg(int i,Stack stack,bool a) const{ return nargs[i] ? GetAny<bool>( (*nargs[i])(stack) ): a;}
  int arg(int i,Stack stack,int a) const{ return nargs[i] ? GetAny<int>( (*nargs[i])(stack) ): a;}
  
  Cemracs2007solver(const basicAC_F0 & args) :nbsol(args.size()-1),sol(args.size()-1)
  {   
      args.SetNameParam(n_name_param,name_param,nargs);
      getmesh=to<pmesh>(args[0]); 
      int ksol=0; 
      ffassert(nbsol<100);
      for (int i=1;i<nbsol+1;i++)       
         if (args[i].left()==atype<E_Array>())
          {
            const E_Array * a = dynamic_cast<const E_Array *>(args[i].LeftValue());
            ffassert(a);
            ksol+=a->size(); 
          }
         else
           ksol++;
      sol.resize(ksol); 
      ksol=0; 
      for (int i=1;i<nbsol+1;i++)       
         if (args[i].left()==atype<E_Array>())
          {
            const E_Array * a = dynamic_cast<const E_Array *>(args[i].LeftValue());
            ffassert(a);
             int N=a->size();
            typesol[i-1]=N-1; // ok en 2D
            if (N<=8) { // used to be 4
                       for (int j=0;j<N;j++)             
                       sol[ksol++]=to<double>((*a)[j]); 
		     }
            else {
             lgerror(" Cemracs2007solver vecteur a plus de 4 compossantes inconnue");
            }
              
          }
         else {
          typesol[i-1]=0;
          sol[ksol++]=to<double>(args[i]);
          }
     const E_Array * expoutput = dynamic_cast<const E_Array *>(nargs[0]);
     if(expoutput)
     {
       if(expoutput->nbitem()!=n_output_param) ExecError("\nSorry we wait an array with 7 components in: output=[vector_triangles(P0) vector_vertices(P1) vector_eta_DF(P1) vector_eta_R(P1)  vector_err_k(P0) vector_err_v(P1) hd(P1) ]");
       for (int i=0;i<n_output_param;i++) em_output[i] = (*expoutput)[i];
     } else
     {
//	cout << "n_output_param " << n_output_param << endl; //" ex... "<< expoutput->nbitem() << endl;
        ExecError("\n CEMRACS_exec:  output=[] is missing ! ");
     }
   }  



    static ArrayOfaType  typeargs() { return  ArrayOfaType(atype<pmesh>(),true);}
    static  E_F0 * f(const basicAC_F0 & args){ return new Cemracs2007solver(args);} 
    AnyType operator()(Stack s) const ;    
    operator aType () const { return atype<pmesh>();}         
};

 basicAC_F0::name_and_type Cemracs2007solver::name_param[Cemracs2007solver::n_name_param] = {
//       {   "medit",         &typeid(bool) }, // 0 do output in medit or not
       {   "output"          ,  &typeid(E_Array)}  // 1 arrays 
//       {   "filename",         &typeid(string*)}, // 2 filename of data to be created (medit)
//       {   "strategy",         &typeid(long) } // 3 adaptation strategy number, see later
    };



// TABOR-20071007
AnyType Cemracs2007solver::operator()(Stack stack) const 
{
  using namespace bamg;
  double pi = bamg::Pi;
  double cutoffradian          = -1.0* pi/180. ;
//  bool medit                   = arg(0,stack,false) ; // use medit for exporting data or not
//  int strategy                 = arg(3,stack,0) ; // use medit for exporting data or not

  KN<double> *mm_output[n_output_param]; for (int i=0;i<n_output_param;i++) mm_output[i]=0;
  string*fn=0;
  if (nargs[2]) fn = GetAny<string*>((*nargs[2])(stack));
  using Fem2D::MeshPoint;
  using Fem2D::Mesh;
  Mesh * Thh = GetAny<pmesh>((*getmesh)(stack));
  ffassert(Thh);
  Triangles * oTh = msh2bamg(*Thh,cutoffradian);
  Triangles &Th(*oTh);

  Th.ReMakeTriangleContainingTheVertex();   // TODO : what does it do ? now idea...
  MeshPoint* mp(Fem2D::MeshPointStack(stack)); // ok - MeshPoint stack is reffered in mp
  long g_nt = Thh->nt; // the number of triangles
  long g_nv = Thh->nv; // the number of vertices

  for (int i=0;i<n_output_param;i++) mm_output[i]= GetAny<KN<double> *>( (*em_output[i])(stack) );
   
//  KN<double> &vector_triangles=*mm_output[0];
  KN<double> &ph=*mm_output[0];
/*  KN<double> &vector_eta_DF=*mm_output[2];
  KN<double> &vector_eta_R=*mm_output[3];
  KN<double> &vector_err_t=*mm_output[4];
  KN<double> &vector_err_v=*mm_output[5];
  KN<double> &vector_h_d=*mm_output[6];
  KN<double> &vector_eta_DF_1=*mm_output[7];
  KN<double> &vector_eta_DF_2=*mm_output[8];*/

  // LIBEREC 2008
  // step 1. inserting the geometry from FF to Martin's code
    double **LIB_V = new double*[g_nv]; 
    int *LIB_V_aux = new int[g_nv];
    double *LIB_R = new double[g_nv];
    for (int i=0;i<g_nv;i++)
    {
       LIB_V[i] = new double[2];  
       LIB_V_aux[i] = 0;
    }
    int **LIB_T = new int*[g_nt]; 
    for (int i=0;i<g_nt;i++) LIB_T[i] = new int[3];  
    
    for (int i=0;i<g_nt;i++)
    {
      for (int j=0;j<3;j++)
      {
         int gvn = Th.Number(Th(i)[j]); // vertex global nmumber
         LIB_T[i][j] = gvn;
	 if (LIB_V_aux[gvn]==0) 
	 { 
	   LIB_V_aux[gvn]=1; 
	   LIB_V[gvn][0] = Th(i)[j].r.x; 
	   LIB_V[gvn][1] = Th(i)[j].r.y; 
	   // now fill out the dirichlet boundary conditions
	   LIB_R[gvn] = reaction(LIB_V[gvn][0],LIB_V[gvn][1],mp,sol,stack);
	   ph[gvn] = ex(LIB_V[gvn][0],LIB_V[gvn][1],mp,sol,stack);
	 }
      }
//      cout << "T"<< i<<" "<<LIB_T[i][0] << " "<<LIB_T[i][1] << " "<<LIB_T[i][2] << endl;
    }

    for (int i=0;i<g_nv;i++)
    {
//      cout << "V"<< i<<" "<<LIB_V[i][0] << " "<<LIB_V[i][1] << endl;
    }
    
// step 2 : determine, if the vertex is on the boundary ------ too complicated because FREE FEM is square root of shit 
   // CEMRACS : go through all vertices and find all adjacent triangles
   int *V_ind = new int[g_nv]; // this is only an indicator used to be sure that we go through all vertices only once
   for (int iv =0;iv<g_nv;iv++) V_ind[iv]=-1;
    // go through all triangles and do algorithms for all vertices
    for (int  it = 0; it < g_nt; it++) 
    {
        bamg::Triangle & Kinit = Th(it); // current triangle we work on
	for (Int4 iv = 0; iv < 3; iv++)
	{
	  int global_vertex_number = Th.Number(Kinit[iv]); // vertex we are messing around
    	  if (V_ind[global_vertex_number]==-1)
	  {
	    // STEP -1 : initialization
  //	    double vx = Kinit[iv].r.x;
//	    double vy = Kinit[iv].r.y;
	    Int4 numKinit = it;               // initial number of the triangle we started from
	    Int4 numKcur =  numKinit;         // current number of the triangle we are in
	    Int4 done =0;                     // is the while loop over ?
	    Int4 boundary = 0; // are we on the boundary ?
	    Int4 numKstart = -1;
	  
	    Int4 num_adj_edge = 0; // number of the edge the next triangle is adjacent to 
	    Int4 local_vertex_num = iv;
	    Int4 n_big_tria = 0;   // number of big triangles around the vertex. 0 = there is no adjacent triangles
	    // STEP 0: now put information that this vertex was already computed
	    // STEP 1a: now go through the triangles  clock wise to find the boundary or simply count the global number of big triangles
	    while (done==0)
	    {
	      // now find the local index of the vertex in the numKcur triangle
	      n_big_tria++;
	      local_vertex_num=-1;
	      for (int i=0;i<3;i++) if (Th.Number(Th(numKcur)[i])==global_vertex_number) local_vertex_num = i;
	      if (local_vertex_num == -1) cout << "CEMRACS error : no match in the local vertex number : the vertex is not in the triangle !!! panic ..." << endl;
	      num_adj_edge = (local_vertex_num+2)%3;  // number of the edge for the adjacent triangle to go : clockwse
	      if (Th.Number(*Th(numKcur).Adj(num_adj_edge)) >= g_nt)  // we reached the edge from the left clock wise direction !!!
	      {
	         boundary = 1;
		 done = 1;
		 numKstart = numKcur;
	      } 
	      numKcur = Th.Number(*Th(numKcur).Adj(num_adj_edge));
	      if (numKcur == numKinit) // hooray, we meet the initial triangle again
	      {
	        done = 1;
		boundary = 0;
		numKstart = numKinit;
	      }
	    } // end wihle done==0
	    V_ind[global_vertex_number] = (boundary) ? 0 : 1;
	} // end if
      }// end for iv
    }// end for it

  // step 2. computing --> --> Martin's code --> ph, grad ph, error (possibly)
/*  for (Int4  iv = 0; iv < g_nv; iv++) 
    {
      ph[iv] = iv*0.1;
    }*/
    
     Martin_solve_us_the_problem(g_nv, g_nt, LIB_V, V_ind, LIB_T,ph, LIB_R);
  // end of LIBEREC 2008
    // step 5 : delete everything
    delete [] V_ind;
    delete [] LIB_V_aux;
    delete []LIB_R;
    for (int i=0;i<g_nv;i++) delete[] LIB_V[i]; delete[] LIB_V;         
    for (int i=0;i<g_nt;i++) delete[] LIB_T[i]; delete[] LIB_T;
  

  double estimator =0;
  return SetAny<double>(estimator);
}  



//   to add the name myfunction to the freefem++ table 
class Init { public:
  Init();
};
Init init;
Init::Init(){
  Global.Add("Cemracs2007solver","(",new OneOperatorCode<Cemracs2007solver>);
}

