#include <iostream.h>
#include <fstream.h>
#include <assert.h>
#include <math.h>
#include "vecteur.cpp"
#include "matrice.cpp"
#include "ligne.cpp"
#include "GC.cpp"
#include "sommet.cpp"
#include "triangle.cpp"
#include "maillage.cpp"
#include "functions.cpp"

#include <stdio.h>

#include "GMMSOLVE.cpp"

//=============================================================================
int Martin_solve_us_the_problem(int ns, int nt, double **V,  int*V_ind, int **T, double*ph, double*R)
// ph ... output solution !
//=============================================================================
  {
  int nsi = 0, i, j, k, l, s[3];
  double sour;
  bool prec = false;
  bool test;

//  ifstream f;
//  f.open ("mesh.txt"); // ouverture du fichier avec le maillage
//  f >> ns >> nt;       // lecture du nombre des sommets et du nombre des triangles

  Maillage M (ns, nt); // construction du maillage

  for (i = 0; i < ns; i++) { // lecture des sommets
//    f >> M[i].x >> M[i].y >> M[i].ind >> M[i].cb; // lecture de la coordonnee x, coordonnee y, indice interieur / bord, condition au bord
    M[i].x = V[i][0];
    M[i].y = V[i][1];
    M[i].ind = V_ind[i]; // indicateur des sommets externes
    M[i].cb = ph[i]; // valeur au bord de la domaine ... dirichlet
//    cout << " hrana : " << ph[i] << endl;
//    cout << "VVV i="<< i<<" x="<<V[i][0] <<" y=" << V[i][1] << " index "<<V_ind[i] << endl;
    if (M[i].ind) {   // pour des sommets internes
      M[i].num = nsi; // numerotation des sommets internes
      nsi++;          // nombre des sommets internes
    }
  }


  for (i = 0; i < nt; i++) { // lecture des triangles
//    f >> s[0] >> s[1] >> s[2] >> M(i).dif >> M(i).sour; // numeros des sommets, coefficient de diffusion, terme source
    for (j=0;j<3;j++) s[j] = T[i][j];
    M(i).dif =1.0;
    M(i).sour=0.0; // la source
    M(i).num = i;            // numerotation des triangles
    for (j = 0; j < 3; j++) {// remplissage des triangles voisins des sommets
      M(i).sommets[j] = &M[s[j]];
      M[s[j]].triangles[M[s[j]].nbtriang] = &M(i);
      M[s[j]].nbtriang++;
    }
    M(i).Geometrie ();       // remplissage des proprietes geometriques des triangles
  }

//  f.close (); // fermeture du fichier avec le maillage

//cout << "nsi is " << nsi << endl;

  Matrice A (nsi, (MAXVOIS + 1) * nsi); // construction de la matrice des elements finis
  Matrice C (nsi, nsi);                 // construction de la matrice de preconditionnement (diagonale)
  Vecteur b (nsi);                      // construction du vecteur du second membre
  Ligne ligneA;                         // une ligne de la matrice A
  Ligne ligneC;                         // une ligne de la matrice C
  EZTriangle* tri;                        // pointeur sur un triangle
  double coef;                          // un coefficient de la matrice / du second membre
  int num;                              // numero du sommet interne
  double react;

  for (i = 0; i < ns; i++) {            // boucle a travers des sommets
    ligneA.Zero();                      // nettoyer le vecteur ligneA
    ligneC.Zero();                      // nettoyer le vecteur ligneC
    sour = 0;                           // nettoyer les sources
    react = R[i]; 			// reaction terme

    if (M[i].ind == 0)                  // sommet au bord du domaine - pas de ligne de A
      continue;

    num = M[i].num;                     // numero du sommet interne

    // reakcni clen    
    double sD=0;
    for (j = 0; j < M[i].nbtriang; j++)  // boucle a travers des triangles du sommet i
      sD+= M[i].triangles[j]->mes;
    
    ligneA[num] += react*sD*11.0/3.0/18.0;
    // konec reakcni clen
    
    
    for (j = 0; j < M[i].nbtriang; j++) { // boucle a travers des triangles du sommet i
      tri = M[i].triangles[j];

      k = 0;
      while (tri->sommets[k] != &M[i])
        k++;  // sommet k du triangle tri correspond au sommet global i

      for (l = 0; l < 3; l++) 
      {
        coef = tri->dif * (tri->fonctbas[l][0] * tri->fonctbas[k][0] + tri->fonctbas[l][1] * tri->fonctbas[k][1]) * tri->mes;
        if (tri->sommets[l]->ind) // sommet interne - entree de la matrice
          ligneA[tri->sommets[l]->num] += coef;
        else                      // sommet externe - valeur connue - entree du second membre
          sour -= coef * tri->sommets[l]->cb;
      } // for l
      
      // reakcni prispevek
      for (l = 0; l < 3; l++)
      {
         if (k!=l) // non-diagonal
	 {
         coef = react* 7.0/6.0/18.0*tri->mes;
        if (tri->sommets[l]->ind) // sommet interne - entree de la matrice
          ligneA[tri->sommets[l]->num] += coef;
        else                      // sommet externe - valeur connue - entree du second membre
          sour -= coef * tri->sommets[l]->cb;
	}

      }
      // konec reakcni prispevek

      sour += tri->mes * tri->sour / 3.0;  // terme source
    } // for j

    if (prec)
//      ligneC[num] = ligneA[num];   // preconditionnement par l'inverse de la diagonale de la matrice A
      ligneC[num] = 1.0 / ligneA[num];   // preconditionnement par l'inverse de la diagonale de la matrice A
    else
      ligneC[num] = 1.0;               // preconditionnement par la matrice d'identite
    A.DefLigne (ligneA, num);          // ligne de la matrice pour le sommet interne num
    C.DefLigne (ligneC, num);          // ligne de la matrice de preconditionnement pour le sommet interne num
    b[num] = sour;                     // second membre pour le sommet interne num
  } // for i

  double eps = 1e-12;              // residu du gradient conjuge preconditionne
  int m = 10*nsi;                    // nombre maximal d'iterations
  int it=-1;                         // nombre actuel d'iterations
  Vecteur x (nsi);                // construction du vecteur de solution
/*  if (test) {                     // impression des matrices A et C et du vecteur b
    cout << A;
    cout << C;
    cout << b;
  }*/
  
//  it = GradientConjugue (A, C, b, x, m, eps);
  double res_gmm;
  size_t nit_gmm;
  solver_gmm(A, b, x, res_gmm, nit_gmm, eps);
  cout << "residu = " << sqrt((A*x-b)*(A*x-b)) << endl
       << "residu gmm = " << res_gmm << " ; #iter gmm = " << nit_gmm << endl;

//   if (it >= 0)
//     cout << "Convergence de la methode du gradient conjugue en " << it << " iterations." << endl;
//   else
//     cout << "Non convergence de la methode du gradient conjugue." << endl;

//  cout << x;

  // conversion du vecteur x 
   for (i=0;i<ns;i++)
   {
//     if (it >=0) // CG a converge
     {
        if (M[i].ind!=0) 
        {
          ph[i] = x[M[i].num];
//          cout << "In. M["<<i<<"].ind="<<M[i].ind<<endl;
	}
        else
        { 
//          ph[i]=0;
          
//          cout << "Out. M["<<i<<"].ind="<<M[i].ind<<endl;
        }
//     } else  {
//        ph[i] = 0;
     }
   }
/*

  double xx[4], yy[4], sl[4], xr, yr, sr;
  double n[2], grad[2];
  int cc;
  double gr1, gr2, cnst, flux;
  double RR, src;


  for (i = 0; i < ns; i++) {            // boucle a travers des sommets

    if (M[i].ind == 0)                  // sommet au bord du domaine - pas de ligne de A
      continue;

    flux = 0.0;
    RR = 0.0;
    src = 0.0;
    num = M[i].num;                     // numero du sommet interne
    react = R[i]; 			// reaction terme


    printf(" Nu du sommet %d \n", num);
    for (j = 0; j < M[i].nbtriang; j++) { // boucle a travers des triangles du sommet i
      tri = M[i].triangles[j];

      k = 0;
      while (tri->sommets[k] != &M[i])
        k++;  // sommet k du triangle tri correspond au sommet global i

      // xx, yy, sl - cordonnees et valeurs dans les sommets
      xx[0] = tri->sommets[k]->x;
      yy[0] = tri->sommets[k]->y;
      sl[0] = x[tri->sommets[k]->num];

      cc = 1;

      for (l = 0; l < 3; l++) {
        if (l==k)
          continue;
        xx[cc] = tri->sommets[l]->x;
        yy[cc] = tri->sommets[l]->y;

        if (tri->sommets[l]->ind == 0)
          sl[cc] = tri->sommets[l]->cb;
        else
          sl[cc] = x[tri->sommets[l]->num];
        cc++;
      } // for l

      CalcGrads (xx[0], yy[0], xx[1], yy[1], xx[2], yy[2], sl[0], sl[1], sl[2], gr1, gr2, cnst);
      grad[0] = gr1;
      grad[1] = gr2;

      // xx, yy, sl - cordonnees et valeurs dans le sommet k, des mi-points d'aretes et du barycentre
      xr = 0.0;
      yr = 0.0;
      sr = 0.0;
      for (l = 0; l < 3; l++) {
        xr += xx[l];
        yr += yy[l];
        sr += sl[l];
      }

      xx[3] = xr / 3.0;
      yy[3] = yr / 3.0;
      sl[3] = sr / 3.0;

      xx[1] = (xx[0] + xx[1]) / 2.0;
      yy[1] = (yy[0] + yy[1]) / 2.0;
      sl[1] = (sl[0] + sl[1]) / 2.0;
      xx[2] = (xx[0] + xx[2]) / 2.0;
      yy[2] = (yy[0] + yy[2]) / 2.0;
      sl[2] = (sl[0] + sl[2]) / 2.0;

      NormVectLine (xx[1], yy[1], xx[3], yy[3], xx[0], yy[0], n);
      flux -= ProdSc (grad, n) * Dist (xx[1], yy[1], xx[3], yy[3]);

      NormVectLine (xx[2], yy[2], xx[3], yy[3], xx[0], yy[0], n);
      flux -= ProdSc (grad, n) * Dist (xx[2], yy[2], xx[3], yy[3]);

      RR += react * (sl[0] + sl[1] + sl[3]) * Surface(xx[0], yy[0], xx[1], yy[1], xx[3], yy[3]) / 3.0;
      RR += react * (sl[0] + sl[2] + sl[3]) * Surface(xx[0], yy[0], xx[2], yy[2], xx[3], yy[3]) / 3.0;

//      printf(" surface %e and %e reaction %e \n", Surface(xx[0], yy[0], xx[1], yy[1], xx[3], yy[3]),Surface(xx[0], yy[0], xx[2], yy[2], xx[3], yy[3]), react);
      printf(" tria : %f,%f  %f,%f  %f,%f (bc)  \n", xx[0],yy[0],xx[1],yy[1],xx[3],yy[3]);
      printf(" values : %f  %f  %f \n", sl[0],sl[1],sl[3]);
      printf(" tria : %f,%f  %f,%f  %f,%f (bc)  \n", xx[0],yy[0],xx[2],yy[2],xx[3],yy[3]);
      printf(" values : %f  %f  %f \n", sl[0],sl[2],sl[3]);
//      printf(" quad : %f,%f  %f,%f  %f,%f bc %f,%f \n", xx[0],yy[0],xx[1],yy[1],xx[2],yy[2],xx[3],yy[3]);
//      printf(" values : %f  %f  %f bc %f \n", sl[0],sl[1],sl[2],sl[3]);
      src += tri->mes * tri->sour / 3.0;
      if (src !=0) printf(" src %e \n",src);

    } // for j

    printf("\n comparing p_h.n_dD %e and rph_D %e diff %e\n", flux, src - RR, flux + RR - src);
  }
*/

  return 0;
  }
