#include <assert.h>
#include "GC.h"

//=============================================================================
int GradientConjugue (const Matrice &A, const Matrice &C, const Vecteur &b,
                      Vecteur &x, int m, double eps)
//=============================================================================
  {
  // verifier si les dimensions sont bonnes
  assert ((A.RetDim() == C.RetDim()) && (C.RetDim() == b.RetDim()) && (b.RetDim() == x.RetDim()));

  int dim = A.RetDim();
  Vecteur g (dim), h (dim), Ah (dim), Cg (dim);
  double g2, g2prev, rho, gamma;
  int iter;

  g = A * x - b;
  Cg = C * g;
  h = Cg * (-1);

  g2 = Cg * g;
  if (fabs (g2) < eps * eps)      // solution deja convergee
    {
    return 0; // OK, convergence
    }

  for (iter = 0; iter < m; iter++)
    {
    Ah = A * h;
    rho = - (g * h) / (h * Ah);
    x += h * rho;
    g += Ah * rho;
    Cg = C * g;
    g2prev = g2;
    g2 = Cg * g;
    if (fabs (g2) < eps * eps)
      {
      return (iter + 1); // OK, convergence
      }

    gamma = g2 / g2prev;
    h *= gamma;
    h -= Cg;
    }

  // non convergence
  return -1;
  }


