#include <math.h>
#include "functions.h"

//=============================================================================
double Dist (double x1, double y1, double x2, double y2)
//=============================================================================
{
  return sqrt ((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
}

//=============================================================================
void NormVectLine (double x1, double y1, double x2, double y2, double x3,
                   double y3, double n[2])
//=============================================================================
{
	double in[2];

	n[0] = (y1 - y2) / Dist(x1, y1, x2, y2);
	n[1] = (x2 - x1) / Dist(x1, y1, x2, y2);
	in[0] = x3 - x1;
	in[1] = y3 - y1;
	if (ProdSc (n, in) > 0.0) {
    n[0] *= -1.0;
    n[1] *= -1.0;
  }
}

//=============================================================================
double ProdSc (double a[2], double b[2])
//=============================================================================
{
  return (a[0] * b[0] + a[1] * b[1]);
}

//=============================================================================
void CalcGrads (double A1, double A2, double B1, double B2, double C1, double C2, double b1,
                     double b2, double b3, double &gr1, double &gr2, double &cnst)
//=============================================================================
{
//gr1, gr2 - composants du gradient d'une fonction linearie(=plan), qui satisfait
//f(A)=b1, f(B)=b2, f(C)=b3, ou A=[A1,A2], B=[B1,B2], C=[C1,C2]
//f(x,y) = gr1*x+gr2*y+cnst

  if (B2 != A2) {
    if (C2 != A2) {
      gr1 =
        (b2 - b1 - (B2 - A2) / (C2 - A2) * (b3 - b1)) / (-(B2 - A2) * (C1 - A1) / (C2 - A2) + B1 -
                                                         A1);
      gr2 = (b2 - b1 - (B1 - A1) * gr1) / (B2 - A2);
      cnst = b1 - A1 * gr1 - A2 * gr2;
    } else {                    // C2 == A2
      gr1 = (b3 - b1) / (C1 - A1);
      gr2 = (b2 - b1 - (B1 - A1) * gr1) / (B2 - A2);
      cnst = b1 - A1 * gr1 - A2 * gr2;
    }
  } else {                      // B2 == A2
    gr1 = (b2 - b1) / (B1 - A1);
    gr2 = (b3 - b1 - (C1 - A1) * gr1) / (C2 - A2);
    cnst = b1 - A1 * gr1 - A2 * gr2;
  }
}

//=============================================================================
double Surface (double A1, double A2, double B1, double B2, double C1, double C2)
//=============================================================================
{
  double u[3], w[3], n[3], surface;

  u[0] = B1 - A1;
  u[1] = B2 - A2;
  u[2] = 0;
  w[0] = C1 - A1;
  w[1] = C2 - A2;
  w[2] = 0;
  ProdVect (u, w, n);

  surface = 0.5 * sqrt (n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);

  return surface;
}

//! Vector pruct.
//=============================================================================
void ProdVect (double u[3], double w[3], double x[3])
//=============================================================================
{
  // u, w - inputs
  // x    - resultat

  x[0] = u[1] * w[2] - u[2] * w[1];
  x[1] = u[2] * w[0] - u[0] * w[2];
  x[2] = u[0] * w[1] - u[1] * w[0];
}

