#include <malloc.h>
#include <assert.h>
#include "matrice.h"

//=============================================================================
Matrice::Matrice (int d, int nb)
//=============================================================================
{
  dim = d;
  nbcoef = nb;
  valeurs = new double [nbcoef]; // allocation dynamique de la memoire
  lignesind = new int [dim + 1]; // allocation dynamique de la memoire
  colonnes = new int [nbcoef];   // allocation dynamique de la memoire
}

//=============================================================================
Matrice::~Matrice ()
//=============================================================================
{
  delete [] valeurs;          // deallocation de la memoire
  valeurs = NULL;             // par messure de securite
  delete [] lignesind;        // deallocation de la memoire
  lignesind = NULL;           // par messure de securite
  delete [] colonnes;         // deallocation de la memoire
  colonnes = NULL;            // par messure de securite
}

//=============================================================================
int Matrice::RetDim () const
//=============================================================================
{ // const - on ne peut pas changer la matrice ici
  return dim;
}

//=============================================================================
Vecteur Matrice::operator * (const Vecteur &V) const
//=============================================================================
{ // const - on ne change pas la matrice
  // const V - on ne change pas V
  assert (dim == V.RetDim()); // memes dimensions
  Vecteur vec (dim);          // on cree un nouveau vecteur
  int i, j;
  for (i = 0; i < dim; i++) {
    vec[i] = 0;
    for (j = lignesind[i]; j < lignesind[i + 1]; j++)
      vec[i] += valeurs[j] * V[colonnes[j]];
  }
  return vec;                 // on retourne le nouveau vecteur
}

//=============================================================================
void Matrice::DefLigne (const Vecteur &ligne, int ind)
//=============================================================================
{
  assert (dim == ligne.RetDim());   // memes dimensions
  assert (ind >= 0 && ind < dim);   // ligne entre 0 et dim
  lignesind[0] = 0;                 // dans nos applications, la premiere valeure non nulle est sur la ligne 0 de la matrice
  int pos = lignesind[ind];         // nombre actuel d'elements non nuls dans la matrice
  for (int i = 0; i < dim; i++) {
    if (ligne[i] != 0) {            // on insere seulement les elements non nuls de V
      colonnes[pos] = i;            // la colonne d'element pos est i
      valeurs[pos] = ligne[i];      // la valeur d'element pos est ligne.valeurs[i]
      pos++; 			     // on a ajoute un nouveau element non nul
    }
  }
  lignesind[ind + 1] = pos;         // nombre actuel d'elements non nuls dans la matrice
}

//=============================================================================
void Matrice::DefLigne (const Ligne &ligne, int ind)
//=============================================================================
{
  assert (ind >= 0 && ind < dim);       // ligne entre 0 et dim
  lignesind[0] = 0;                     // dans nos applications, la premiere valeure non nulle est sur la ligne 0 de la matrice
  int pos = lignesind[ind];             // nombre actuel d'elements non nuls dans la matrice
  for (int i = 0;i < dim;i++)
  for (int j=0 ; j < ligne.nbcoef;j++)
  {
    if (i==ligne.colonnes[j])
    {
    colonnes[pos] = ligne.colonnes[j];  // la colonne d'element pos est i
    valeurs[pos] = ligne.valeurs[j];    // la valeur d'element pos est ligne.valeurs[i]
    pos++;
    }                              // on a ajoute un nouveau element non nul
  }
//  for (int i = 0; i < ligne.nbcoef; i++) {
//    colonnes[pos] = ligne.colonnes[i];  // la colonne d'element pos est i
//    valeurs[pos] = ligne.valeurs[i];    // la valeur d'element pos est ligne.valeurs[i]
//    pos++;                              // on a ajoute un nouveau element non nul
//  }
//tohle bylo cele blbe, protoze pro umfpack je potreba mit indexy sloupcu srovnane podle velikosti
  lignesind[ind + 1] = pos;             // nombre actuel d'elements non nuls dans la matrice
}

//=============================================================================
ostream & operator << (ostream &os, const Matrice &M)
//=============================================================================
{
  int i, j;
  for (i = 0; i < M.dim; i++) {
    for (j = M.lignesind[i]; j < M.lignesind[i+1]; j++)
      os << "ligne " << i << " colonne " << M.colonnes[j] << " : " << M.valeurs[j] << endl;
  }
  return os;
}

