#include <malloc.h>
#include <assert.h>
#include "vecteur.h"

//=============================================================================
Vecteur::Vecteur (int d)
//=============================================================================
{
  dim = d;
  valeurs = new double [dim]; // allocation dynamique de la memoire
}

//=============================================================================
Vecteur::Vecteur (const Vecteur &V)
//=============================================================================
{
  dim = V.dim;
  valeurs = new double [dim]; // allocation dynamique de la memoire
  for (int i = 0; i < dim; i++)
    valeurs[i] = V.valeurs[i];
}

//=============================================================================
Vecteur::~Vecteur ()
//=============================================================================
{
  delete [] valeurs;          // deallocation de la memoire
  valeurs = NULL;             // par messure de securite
}

//=============================================================================
int Vecteur::RetDim () const
//=============================================================================
{ // const - on ne peut pas changer le vecteur ici
  return dim;
}

//=============================================================================
double Vecteur::operator [] (int i) const
//=============================================================================
{ // const - on ne peut pas changer le vecteur ici
  assert (i >= 0 && i < dim); // arret si i < 0 ou i >= dim
  return valeurs[i];
}

//=============================================================================
double & Vecteur::operator [] (int i)
//=============================================================================
{ // double & - on va changer des elements du vecteur
  assert (i >= 0 && i < dim); // arret si i < 0 ou i >= dim
  return valeurs[i];
}

//=============================================================================
Vecteur & Vecteur::operator = (const Vecteur &V)
//=============================================================================
{ // Vecteur & - on va changer le vecteur
  // const V - on ne change pas V
  assert (V.dim == dim); // memes dimensions
  for (int i = 0; i < dim; i++)
    valeurs[i] = V.valeurs[i];
  return *this;
}

//=============================================================================
Vecteur Vecteur::operator + (const Vecteur &V) const
//=============================================================================
{ // const - on ne change pas le vecteur
  // const V - on ne change pas V
  assert (V.dim == dim); // memes dimensions
  Vecteur vec (dim);     // on cree un nouveau vecteur
  for (int i = 0; i < dim; i++)
    vec.valeurs[i] = valeurs[i] + V.valeurs[i];
  return vec;            // on retourne le nouveau vecteur
}

//=============================================================================
Vecteur Vecteur::operator - (const Vecteur &V) const
//=============================================================================
{ // const - on ne change pas le vecteur
  // const V - on ne change pas V
  assert (V.dim == dim); // memes dimensions
  Vecteur vec (dim);     // on cree un nouveau vecteur
  for (int i = 0; i < dim; i++)
    vec.valeurs[i] = valeurs[i] - V.valeurs[i];
  return vec;            // on retourne le nouveau vecteur
}

//=============================================================================
Vecteur & Vecteur::operator += (const Vecteur &V)
//=============================================================================
{ // Vecteur & - on va changer le vecteur
  // const V - on ne change pas V
  assert (V.dim == dim); // memes dimensions
  for (int i = 0; i < dim; i++)
    valeurs[i] += V.valeurs[i];
  return *this;
}

//=============================================================================
Vecteur & Vecteur::operator -= (const Vecteur &V)
//=============================================================================
{ // Vecteur & - on va changer le vecteur
  // const V - on ne change pas V
  assert (V.dim == dim); // memes dimensions
  for (int i = 0; i < dim; i++)
    valeurs[i] -= V.valeurs[i];
  return *this;
}

//=============================================================================
Vecteur Vecteur::operator * (double alpha) const
//=============================================================================
{ // const - on ne change pas le vecteur
  Vecteur vec (dim);     // on cree un nouveau vecteur
  for (int i = 0; i < dim; i++)
    vec.valeurs[i] = valeurs[i] * alpha;
  return vec;            // on retourne le nouveau vecteur
}

//=============================================================================
Vecteur & Vecteur::operator *= (double alpha)
//=============================================================================
{ // Vecteur & - on va changer le vecteur
  for (int i = 0; i < dim; i++)
    valeurs[i] *= alpha;
  return *this;
}

//=============================================================================
double Vecteur::operator * (const Vecteur &V) const
//=============================================================================
{ // const - on ne change pas le vecteur
  // const V - on ne change pas V
  assert (V.dim == dim); // memes dimensions
  double res = 0.0;
  for (int i = 0; i < dim; i++)
    res += valeurs[i] * V.valeurs[i];
  return res;
}

//=============================================================================
ostream & operator << (ostream &os, const Vecteur &V)
//=============================================================================
{
  for (int i = 0; i < V.dim; i++)
    os << V.valeurs[i] << " ";
  os << endl;
  return os;
}

