#ifndef VECTEUR_H
#define VECTEUR_H

#include <iostream.h>

//! Classe d'un vecteur reel.
class Vecteur {

  int dim;              //!< Dimension du vecteur.

public:
  double* valeurs;      //!< Valeurs du vecteur.

  //! Constructeur.
  Vecteur (int d);

  //! Constructeur par copie.
  Vecteur (const Vecteur &V);

  //! Destructeur.
  ~Vecteur ();

  //! Retourne dimension du vecteur.
  int RetDim() const;

  //! Operateur d'acces aux elements du vecteur.
  double operator [] (int i) const;

  //! Operateur de changement des elements du vecteur.
  double & operator [] (int i);

  //! Operateur d'assignement.
  Vecteur & operator = (const Vecteur &V);

  //! Somme de deux vecteurs.
  Vecteur operator + (const Vecteur &V) const;

  //! Soustraction de deux vecteurs.
  Vecteur operator - (const Vecteur &V) const;

  //! Addition d'un vecteur.
  Vecteur & operator += (const Vecteur &V);

  //! Soustraction d'un vecteur.
  Vecteur & operator -= (const Vecteur &V);

  //! Multiplication par un scalaire.
  Vecteur operator * (double alpha) const;

  //! Multiplication par un scalaire.
  Vecteur & operator *= (double alpha);

  //! Produit scalaire de deux vecteurs.
  double operator * (const Vecteur &V) const;

  //! Ecrire le vecteur.
  friend ostream & operator << (ostream &os, const Vecteur &V);
};


#endif
